# Copyright 2021 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import re

# API constants
API_REF_URL = 'https://docs.openstack.org/api-ref/dns'
CURRENT = 'CURRENT'
SUPPORTED = 'SUPPORTED'
DEPRECATED = 'DEPRECATED'
EXPERIMENTAL = 'EXPERIMENTAL'

# RBAC related constants
RBAC_PROJECT_ID = 'project_id'
RBAC_TARGET_PROJECT_ID = 'target_project_id'

# Statuses
ACTIVE = 'ACTIVE'
ERROR = 'ERROR'
INACTIVE = 'INACTIVE'
PENDING = 'PENDING'

# Actions
CREATE = 'CREATE'
DELETE = 'DELETE'
NONE = 'NONE'
UPDATE = 'UPDATE'

# Floating IP constants
FLOATING_IP_ACTIONS = [CREATE, DELETE, UPDATE, NONE]
FLOATING_IP_STATUSES = [ACTIVE, ERROR, INACTIVE, PENDING]

# Quotas
MIN_QUOTA = -1
MAX_QUOTA = 2147483647
QUOTA_API_EXPORT_SIZE = 'api_export_size'
QUOTA_RECORDSET_RECORDS = 'recordset_records'
QUOTA_ZONE_RECORDS = 'zone_records'
QUOTA_ZONE_RECORDSETS = 'zone_recordsets'
QUOTA_ZONES = 'zones'
VALID_QUOTAS = [QUOTA_API_EXPORT_SIZE, QUOTA_RECORDSET_RECORDS,
                QUOTA_ZONE_RECORDS, QUOTA_ZONE_RECORDSETS, QUOTA_ZONES]

# RBAC scopes
PROJECT = 'project'

# Zone constants
ZONE_PRIMARY = 'PRIMARY'
ZONE_SECONDARY = 'SECONDARY'
ZONE_CATALOG = 'CATALOG'
ZONE_TYPES = [ZONE_PRIMARY, ZONE_SECONDARY, ZONE_CATALOG]

# Record regexes
RE_HOSTNAME = re.compile(r'^(?!.{255,})(?:(?:^\*|(?!\-)[A-Za-z0-9_\-]{1,63})(?<!\-)\.)+\Z')  # noqa
RE_ZONENAME = re.compile(r'^(?!.{255,})(?:(?!\-)[A-Za-z0-9_\-]{1,63}(?<!\-)\.)+\Z')  # noqa
RE_SRV_HOST_NAME = re.compile(r'^(?:(?!\-)(?:\_[A-Za-z0-9_\-]{1,63}\.){2})(?!.{255,})(?:(?!\-)[A-Za-z0-9_\-]{1,63}(?<!\-)\.)+\Z')  # noqa
RE_SSHFP_FINGERPRINT = re.compile(r'^([0-9A-Fa-f]{10,40}|[0-9A-Fa-f]{64})\Z')
RE_TLDNAME = re.compile(r'^(?!.{255,})(?:(?!\-)[A-Za-z0-9_\-]{1,63}(?<!\-))(?:\.(?:(?!\-)[A-Za-z0-9_\-]{1,63}(?<!\-)))*\Z')  # noqa
RE_NAPTR_FLAGS = re.compile(r'^[APSUapsu]*$')
RE_NAPTR_SERVICE = re.compile(r'^([A-Za-z]([A-Za-z0-9]*)(\+[A-Za-z]([A-Za-z0-9]{0,31}))*)?$')  # noqa
RE_NAPTR_REGEXP = re.compile(r'^(([^0-9i\\])(.*)\2((.+)|(\\[1-9]))?\2(i?))?$')
RE_KVP = re.compile(r'^\s[A-Za-z0-9]+=[A-Za-z0-9]+')
RE_URL_MAIL = re.compile(r'^mailto:[A-Za-z0-9_\-]+(\+[A-Za-z0-9_\-]+)?@.*')
RE_URL_HTTP = re.compile(r'^http(s)?://.*/')
RE_CERT_TYPE = re.compile(r'(^[A-Z]+$)|(^[0-9]+$)')
RE_CERT_ALGO = re.compile(r'(^[A-Z]+[A-Z0-9\-]+[A-Z0-9]$)|(^[0-9]+$)')

# Floating IP regexes
RE_FIP = re.compile(r'^(?P<region>[A-Za-z0-9\.\-_]{1,100}):(?P<id>[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12})$')  # noqa

# Error Validation regexes
RE_REQUIRED = re.compile(r'\'([\w]*)\' is a required property')

TSIG_ALGORITHMS = [
                'hmac-md5',
                'hmac-sha1',
                'hmac-sha224',
                'hmac-sha256',
                'hmac-sha384',
                'hmac-sha512'
            ]



# Copyright 2021 NS1 Inc. https://www.ns1.com
#
# Author: Dragan Blagojevic <dblagojevic@daitan.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from oslo_log import log as logging
import requests

from designate.backend import base
import designate.conf
from designate import exceptions


LOG = logging.getLogger(__name__)
CONF = designate.conf.CONF


class NS1Backend(base.Backend):
    __plugin_name__ = 'ns1'

    __backend_status__ = 'untested'

    def __init__(self, target):
        super().__init__(target)

        self.api_endpoint = 'https://' + self.options.get('api_endpoint')
        self.api_token = self.options.get('api_token')
        self.tsigkey_name = self.options.get('tsigkey_name', None)
        self.tsigkey_hash = self.options.get('tsigkey_hash', None)
        self.tsigkey_value = self.options.get('tsigkey_value', None)

        self.headers = {
            'X-NSONE-Key': self.api_token
        }

    def _build_url(self, zone):
        return '{}/v1/zones/{}'.format(
            self.api_endpoint, zone.name.rstrip('.')
        )

    def _get_master(self):
        try:
            return self.masters[0]
        except IndexError as e:
            LOG.error('No masters host set in pools.yaml')
            raise exceptions.Backend(e)

    def _check_zone_exists(self, zone):
        try:
            requests.get(
                self._build_url(zone),
                headers=self.headers
            ).raise_for_status()
        except requests.HTTPError as e:
            if e.response.status_code == 404:
                return False
            else:
                LOG.error('HTTP error in check zone exists. Zone %s', zone)
                raise exceptions.Backend(e)
        except requests.ConnectionError as e:
            LOG.error('Connection error in check zone exists. Zone %s', zone)
            raise exceptions.Backend(e)

        return True

    def create_zone(self, context, zone):
        master = self._get_master()
        # designate requires "." at end of zone name, NS1 requires omitting
        data = {
            'zone': zone.name.rstrip('.'),
            'secondary': {
                'enabled': True,
                'primary_ip': master.host,
                'primary_port': master.port
            }
        }
        if self.tsigkey_name:
            tsig = {
                'enabled': True,
                'hash': self.tsigkey_hash,
                'name': self.tsigkey_name,
                'key': self.tsigkey_value
            }
            data['secondary']['tsig'] = tsig

        if not self._check_zone_exists(zone):
            try:
                requests.put(
                    self._build_url(zone),
                    json=data,
                    headers=self.headers
                ).raise_for_status()
            except requests.HTTPError as e:
                # check if the zone was actually created
                if self._check_zone_exists(zone):
                    LOG.info(
                        '%s was created with an error. Deleting zone',
                        zone.name
                    )
                    try:
                        self.delete_zone(context, zone)
                    except exceptions.Backend:
                        LOG.error(
                            'Could not delete errored zone %s', zone.name
                        )
                raise exceptions.Backend(e)
        else:
            LOG.info(
                "Can't create zone %s because it already exists", zone.name
            )

    def delete_zone(self, context, zone, zone_params=None):
        """Delete a DNS zone"""

        # First verify that the zone exists
        if self._check_zone_exists(zone):
            try:
                requests.delete(
                    self._build_url(zone),
                    headers=self.headers
                ).raise_for_status()
            except requests.HTTPError as e:
                raise exceptions.Backend(e)
        else:
            LOG.warning(
                'Trying to delete zone %s but that zone is not present in the '
                'ns1 backend. Assuming success.', zone
            )



# Copyright 2023 inovex GmbH
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from designate import objects
from designate.objects.adapters.yaml import base


class PoolCatalogZoneYAMLAdapter(base.YAMLAdapter):
    ADAPTER_OBJECT = objects.PoolCatalogZone
    MODIFICATIONS = {
        'fields': {
            'catalog_zone_fqdn': {
                'read_only': False
            },
            'catalog_zone_refresh': {
                'read_only': False
            },
            'catalog_zone_tsig_key': {
                'read_only': False
            },
            'catalog_zone_tsig_algorithm': {
                'read_only': False
            },
        }
    }



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from unittest import mock

from oslo_config import fixture as cfg_fixture
import oslotest.base

from designate.cmd import api
from designate.cmd import central
from designate.cmd import mdns
from designate.cmd import producer
from designate.cmd import sink
from designate.cmd import worker
import designate.conf


CONF = designate.conf.CONF


@mock.patch('designate.service.wait')
@mock.patch('designate.service.serve')
@mock.patch('designate.heartbeat_emitter.get_heartbeat_emitter')
@mock.patch('oslo_log.log.setup')
@mock.patch('designate.utils.read_config')
class CmdTestCase(oslotest.base.BaseTestCase):
    def setUp(self):
        super().setUp()
        self.useFixture(cfg_fixture.Config(CONF))

    @mock.patch('designate.api.service.Service')
    def test_api(self, mock_service, mock_read_config, mock_log_setup,
                 mock_heartbeat, mock_serve, mock_wait):
        CONF.set_override('workers', 1, 'service:api')

        api.main()

        mock_read_config.assert_called_with('designate', mock.ANY)
        mock_log_setup.assert_called_with(mock.ANY, 'designate')
        mock_service.assert_called_with()
        mock_heartbeat.assert_called()
        mock_serve.assert_called_with(mock.ANY, workers=1)
        mock_wait.assert_called_with()

    @mock.patch('designate.central.service.Service')
    def test_central(self, mock_service, mock_read_config, mock_log_setup,
                     mock_heartbeat, mock_serve, mock_wait):
        CONF.set_override('workers', 1, 'service:central')

        central.main()

        mock_read_config.assert_called_with('designate', mock.ANY)
        mock_log_setup.assert_called_with(mock.ANY, 'designate')
        mock_service.assert_called_with()
        mock_heartbeat.assert_called()
        mock_serve.assert_called_with(mock.ANY, workers=1)
        mock_wait.assert_called_with()

    @mock.patch('designate.mdns.service.Service')
    def test_mdns(self, mock_service, mock_read_config, mock_log_setup,
                  mock_heartbeat, mock_serve, mock_wait):
        CONF.set_override('workers', 1, 'service:mdns')

        mdns.main()

        mock_read_config.assert_called_with('designate', mock.ANY)
        mock_log_setup.assert_called_with(mock.ANY, 'designate')
        mock_service.assert_called_with()
        mock_heartbeat.assert_called()
        mock_serve.assert_called_with(mock.ANY, workers=1)
        mock_wait.assert_called_with()

    @mock.patch('designate.producer.service.Service')
    def test_producer(self, mock_service, mock_read_config, mock_log_setup,
                      mock_heartbeat, mock_serve, mock_wait):
        CONF.set_override('workers', 1, 'service:producer')

        producer.main()

        mock_read_config.assert_called_with('designate', mock.ANY)
        mock_log_setup.assert_called_with(mock.ANY, 'designate')
        mock_service.assert_called_with()
        mock_heartbeat.assert_called()
        mock_serve.assert_called_with(mock.ANY, workers=1)
        mock_wait.assert_called_with()

    @mock.patch('designate.sink.service.Service')
    def test_sink(self, mock_service, mock_read_config, mock_log_setup,
                  mock_heartbeat, mock_serve, mock_wait):
        CONF.set_override('workers', 1, 'service:sink')

        sink.main()

        mock_read_config.assert_called_with('designate', mock.ANY)
        mock_log_setup.assert_called_with(mock.ANY, 'designate')
        mock_service.assert_called_with()
        mock_heartbeat.assert_called()
        mock_serve.assert_called_with(mock.ANY, workers=1)
        mock_wait.assert_called_with()

    @mock.patch('designate.worker.service.Service')
    def test_worker(self, mock_service, mock_read_config, mock_log_setup,
                    mock_heartbeat, mock_serve, mock_wait):
        CONF.set_override('workers', 1, 'service:worker')

        worker.main()

        mock_read_config.assert_called_with('designate', mock.ANY)
        mock_log_setup.assert_called_with(mock.ANY, 'designate')
        mock_service.assert_called_with()
        mock_heartbeat.assert_called()
        mock_serve.assert_called_with(mock.ANY, workers=1)
        mock_wait.assert_called_with()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import oslotest.base

from designate import objects


class ZoneMasterTest(oslotest.base.BaseTestCase):
    def test_zone_master_from_data(self):
        zone_master = objects.ZoneMaster.from_data('192.0.2.2:5354')
        self.assertEqual('192.0.2.2', zone_master.host)
        self.assertEqual(5354, zone_master.port)

    def test_zone_master_to_data(self):
        zone_master = objects.ZoneMaster.from_data('192.0.2.2:5354')

        self.assertEqual('192.0.2.2:5354', zone_master.to_data())

    def test_zone_masters_from_list(self):
        zone_masters = objects.ZoneMasterList.from_list([
            {'host': '192.0.2.1', 'port': 5354, },
            {'host': '192.0.2.2', 'port': 5354, },
            {'host': '192.0.2.100', 'port': 5354, },
        ])

        self.assertEqual(3, len(zone_masters))

        for zone_master in zone_masters:
            self.assertIn('192.0.2.', zone_master.host)
            self.assertEqual(5354, zone_master.port)

    def test_zone_masters_to_list(self):
        zone_masters_payload = [
            {'host': '192.0.2.1', 'port': 53, },
            {'host': '192.0.2.100', 'port': 5354, },
        ]

        zone_masters = objects.ZoneMasterList.from_list(zone_masters_payload)

        self.assertEqual(2, len(zone_masters))

        for zone_master in zone_masters.to_list():
            self.assertIn(zone_master, zone_masters_payload)

    def test_zone_masters_to_data(self):
        expected_results = [
            '192.0.2.1:53',
            '192.0.2.2:5354'
        ]

        zone_masters = [
            {'host': '192.0.2.1', 'port': 53, },
            {'host': '192.0.2.2', 'port': 5354, },
        ]

        zone_masters = objects.ZoneMasterList.from_list(zone_masters)

        self.assertEqual(2, len(zone_masters))

        for zone_master in zone_masters.to_data():
            self.assertIn(zone_master, expected_results)



# Copyright 2016 Rackspace Inc.
#
# Author: Tim Simmons <tim.simmons@rackspace.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from oslo_log import log as logging
import oslo_messaging as messaging

from designate.common.decorators import rpc as rpc_decorator
from designate.common import profiler
import designate.conf
from designate import rpc


CONF = designate.conf.CONF
LOG = logging.getLogger(__name__)

WORKER_API = None


@profiler.trace_cls("rpc")
@rpc_decorator.rpc_logging(LOG, 'worker')
class WorkerAPI:
    """
    Client side of the worker RPC API.

    API version history:

        1.0 - Initial version
        1.1 - Added perform_zone_xfr and get_serial_number
        1.2 - Added hard_delete to delete_zone
    """
    RPC_API_VERSION = '1.2'

    def __init__(self, topic=None):
        self.topic = topic if topic else CONF['service:worker'].topic

        target = messaging.Target(topic=self.topic,
                                  version=self.RPC_API_VERSION)
        self.client = rpc.get_client(target, version_cap='1.2')

    @classmethod
    def get_instance(cls):
        """
        The rpc.get_client() which is called upon the API object initialization
        will cause a assertion error if the designate.rpc.TRANSPORT isn't setup
        by rpc.init() before.

        This fixes that by creating the rpcapi when demanded.
        """
        global WORKER_API
        if not WORKER_API:
            WORKER_API = cls()
        return WORKER_API

    def create_zone(self, context, zone):
        return self.client.cast(
            context, 'create_zone', zone=zone)

    def update_zone(self, context, zone):
        return self.client.cast(
            context, 'update_zone', zone=zone)

    def delete_zone(self, context, zone, hard_delete=False):
        return self.client.cast(
            context, 'delete_zone', zone=zone, hard_delete=hard_delete)

    def recover_shard(self, context, begin, end):
        return self.client.cast(
            context, 'recover_shard', begin=begin, end=end)

    def start_zone_export(self, context, zone, export):
        return self.client.cast(
            context, 'start_zone_export', zone=zone, export=export)

    def perform_zone_xfr(self, context, zone, servers=None):
        return self.client.cast(
            context, 'perform_zone_xfr', zone=zone, servers=servers)

    def get_serial_number(self, context, zone, host, port):
        return self.client.call(
            context, 'get_serial_number', zone=zone,
            host=host, port=port,
        )



# Copyright 2014 Hewlett-Packard Development Company, L.P.
#
# Author: Endre Karlson <endre.karlson@hpe.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import dns.exception
import dns.message
import dns.opcode
import dns.rcode
import dns.rdatatype
import dns.renderer
import dns.tsig
from oslo_log import log as logging

import designate.conf
from designate import context
from designate import exceptions

CONF = designate.conf.CONF
LOG = logging.getLogger(__name__)


class DNSMiddleware:
    """Base DNS Middleware class with some utility methods"""
    def __init__(self, application):
        self.application = application

    def process_request(self, request):
        """Called on each request.

        If this returns None, the next application down the stack will be
        executed. If it returns a response then that response will be returned
        and execution will stop here.
        """
        return None

    def process_response(self, response):
        """Do whatever you'd like to the response."""
        return response

    def __call__(self, request):
        response = self.process_request(request)

        if response:
            return response

        response = self.application(request)
        return self.process_response(response)

    def _build_error_response(self):
        response = dns.message.make_response(
            dns.message.make_query('unknown', dns.rdatatype.A))
        response.set_rcode(dns.rcode.FORMERR)

        return response


class SerializationMiddleware(DNSMiddleware):
    """DNS Middleware to serialize/deserialize DNS Packets"""

    def __init__(self, application, tsig_keyring=None):
        super().__init__(application)
        self.tsig_keyring = tsig_keyring

    def __call__(self, request):
        # Generate the initial context. This may be updated by other middleware
        # as we learn more information about the Request.
        ctxt = context.DesignateContext.get_admin_context(all_tenants=True)

        message = None
        try:
            message = dns.message.from_wire(request['payload'],
                                            self.tsig_keyring)

            if message.had_tsig:
                LOG.debug('Request signed with TSIG key: %s', message.keyname)

            # Create + Attach the initial "environ" dict. This is similar to
            # the environ dict used in typical WSGI middleware.
            message.environ = {
                'context': ctxt,
                'addr': request['addr'],
            }
        except dns.message.UnknownTSIGKey:
            LOG.error(
                'Unknown TSIG key from %(host)s:%(port)d',
                {
                    'host': request['addr'][0],
                    'port': request['addr'][1]
                }
            )
        except dns.tsig.BadSignature:
            LOG.error(
                'Invalid TSIG signature from %(host)s:%(port)d',
                {
                    'host': request['addr'][0],
                    'port': request['addr'][1]
                }
            )
        except dns.exception.DNSException:
            LOG.error(
                'Failed to deserialize packet from %(host)s:%(port)d',
                {
                    'host': request['addr'][0],
                    'port': request['addr'][1]
                }
            )
        except Exception:
            LOG.exception(
                'Unknown exception deserializing packet '
                'from %(host)s %(port)d',
                {
                    'host': request['addr'][0],
                    'port': request['addr'][1]
                }
            )

        if message is None:
            # NOTE(eandersson): Unsure on the intent of the error handling
            #                   in this code. Cleaning the code path up, but
            #                   leaving functionality as it was.
            # error_response = self._build_error_response()
            # yield error_response.to_wire()
            yield
            return

        # Hand the Deserialized packet onto the Application
        for response in self.application(message):
            # Serialize and return the response if present
            if isinstance(response, dns.message.Message):
                yield response.to_wire(max_size=65535)
            elif isinstance(response, dns.renderer.Renderer):
                yield response.get_wire()
            else:
                LOG.error('Unexpected response %r', response)


class TsigInfoMiddleware(DNSMiddleware):
    """Middleware which looks up the information available for a TsigKey"""

    def __init__(self, application, storage):
        super().__init__(application)
        self.storage = storage

    def process_request(self, request):
        if not request.had_tsig:
            return None

        try:
            name = request.keyname.to_text(True)
            criterion = {'name': name}
            tsigkey = self.storage.find_tsigkey(
                context.get_current(), criterion
            )

            request.environ['tsigkey'] = tsigkey
            request.environ['context'].tsigkey_id = tsigkey.id

        except exceptions.TsigKeyNotFound:
            # This should never happen, as we just validated the key. Except
            # for race conditions.
            return self._build_error_response()

        return None



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import versionutils
from oslo_policy import policy


DEPRECATED_REASON = """
The designate API now supports system scope and default roles.
"""

RULE_ANY = "@"

# Generic policy check string for system administrators. These are the people
# who need the highest level of authorization to operate the deployment.
# They're allowed to create, read, update, or delete any system-specific
# resource. They can also operate on project-specific resources where
# applicable (e.g., cleaning up blacklists)
SYSTEM_ADMIN = 'role:admin'

# Generic policy check string for read-only access to system-level resources.
# This persona is useful for someone who needs access for auditing or even
# support. These uses are also able to view project-specific resources where
# applicable (e.g., listing all pools)
SYSTEM_READER = 'role:admin'

# This check string is reserved for actions that require the highest level of
# authorization on a project or resources within the project
PROJECT_ADMIN = 'role:admin and project_id:%(project_id)s'

# This check string is the primary use case for typical end-users, who are
# working with resources that belong to a project (e.g., creating DNS zones)
PROJECT_MEMBER = 'role:member and project_id:%(project_id)s'

# This check string should only be used to protect read-only project-specific
# resources. It should not be used to protect APIs that make writable changes.
PROJECT_READER = 'role:reader and project_id:%(project_id)s'

# The following are common composite check strings that are useful for
# protecting APIs designed to operate with multiple scopes
SYSTEM_ADMIN_OR_PROJECT_MEMBER = (
    '(' + SYSTEM_ADMIN + ') or (' + PROJECT_MEMBER + ')'
)
SYSTEM_OR_PROJECT_READER = (
    '(' + SYSTEM_READER + ') or (' + PROJECT_READER + ')'
)

# Designate specific "secure RBAC" rules
ALL_TENANTS = 'True:%(all_tenants)s'

ALL_TENANTS_READER = ALL_TENANTS + ' and role:reader'

SYSTEM_OR_PROJECT_READER_OR_ALL_TENANTS_READER = (
    '(' + SYSTEM_READER + ') or (' + PROJECT_READER + ') or (' +
    ALL_TENANTS_READER + ')'
)

SYSTEM_OR_PROJECT_READER_OR_SHARED = (
        SYSTEM_OR_PROJECT_READER + ' or (\'True\':%(zone_shared)s)'
)

RULE_ZONE_TRANSFER = (
    '(' + SYSTEM_ADMIN_OR_PROJECT_MEMBER + ') or '
    'project_id:%(target_project_id)s or '
    'None:%(target_project_id)s')


# Deprecated in Wallaby as part of the "secure RBAC" work.
# TODO(johnsom) remove when the deprecated RBAC rules are removed.
RULE_ADMIN = 'rule:admin'
RULE_ADMIN_OR_OWNER = 'rule:admin_or_owner'
LEGACY_RULE_ZONE_TRANSFER = (
    "rule:admin_or_owner OR "
    "project_id:%(target_tenant_id)s "
    "OR None:%(target_tenant_id)s"
)
RULE_ADMIN_OR_OWNER_OR_SHARED = (
        RULE_ADMIN_OR_OWNER + ' or (\'True\':%(zone_shared)s)'
)

deprecated_default = policy.DeprecatedRule(
    name="default",
    check_str=RULE_ADMIN_OR_OWNER,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)

rules = [
    # TODO(johnsom) remove when the deprecated RBAC rules are removed.
    policy.RuleDefault(
        name="admin",
        check_str="role:admin or is_admin:True"),
    # TODO(johnsom) remove when the deprecated RBAC rules are removed.
    policy.RuleDefault(
        name="owner",
        check_str="project_id:%(tenant_id)s"),
    # TODO(johnsom) remove when the deprecated RBAC rules are removed.
    policy.RuleDefault(
        name="admin_or_owner",
        check_str="rule:admin or rule:owner"),

    # Default policy
    policy.RuleDefault(
        name="default",
        check_str=SYSTEM_ADMIN_OR_PROJECT_MEMBER,
        deprecated_rule=deprecated_default),
]


def list_rules():
    return rules



# Copyright 2013 Hewlett-Packard Development Company, L.P.
#
# Author: Kiall Mac Innes <kiall@managedit.ie>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import designate.conf
from designate.tests.functional.api import v2


CONF = designate.conf.CONF


class ApiV2LimitsTest(v2.ApiV2TestCase):
    def test_get_limits(self):
        response = self.client.get('/limits/')

        self.assertEqual(200, response.status_int)
        self.assertEqual('application/json', response.content_type)

        self.assertIn('max_zones', response.json)
        self.assertIn('max_zone_records', response.json)
        self.assertIn('max_zone_recordsets',
                      response.json)
        self.assertIn('max_recordset_records',
                      response.json)
        self.assertIn('min_ttl', response.json)
        self.assertIn('max_zone_name_length',
                      response.json)
        self.assertIn('max_recordset_name_length',
                      response.json)
        self.assertIn('max_page_limit',
                      response.json)

        absolutelimits = response.json

        self.assertEqual(CONF.quota_zones, absolutelimits['max_zones'])
        self.assertEqual(CONF.quota_zone_records,
                         absolutelimits['max_zone_recordsets'])
        self.assertEqual(CONF['service:central'].min_ttl,
                         absolutelimits['min_ttl'])
        self.assertEqual(CONF['service:central'].max_zone_name_len,
                         absolutelimits['max_zone_name_length'])
        self.assertEqual(CONF['service:central'].max_recordset_name_len,
                         absolutelimits['max_recordset_name_length'])
        self.assertEqual(CONF['service:api'].max_limit_v2,
                         absolutelimits['max_page_limit'])



# Copyright 2012 Managed I.T.
#
# Author: Kiall Mac Innes <kiall@managedit.ie>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

import oslotest.base

from designate import context
from designate import policy


class TestDesignateContext(oslotest.base.BaseTestCase):
    def test_tsigkey_id_override(self):
        orig = context.DesignateContext(
            tsigkey_id='12345', project_id='54321'
        )
        copy = orig.to_dict()

        self.assertEqual('TSIG:12345 54321 - - -', copy['user_identity'])

    @mock.patch.object(policy, 'check')
    def test_sudo(self, mock_policy_check):
        ctxt = context.DesignateContext(
            user_id='12345', project_id='old_project'
        )
        ctxt.sudo('new_project')

        self.assertTrue(mock_policy_check.called)
        self.assertEqual('new_project', ctxt.project_id)
        self.assertEqual('old_project', ctxt.original_project_id)

    def test_get_auth_plugin(self):
        ctx = context.DesignateContext()
        self.assertIsInstance(
            ctx.get_auth_plugin(), context._ContextAuthPlugin
        )

    @mock.patch('keystoneauth1.access.service_catalog.ServiceCatalogV2')
    def test_get_auth_plugin_get_endpoint(self, mock_sc):
        mock_session = mock.Mock()
        mock_service_catalog = mock.Mock()
        mock_sc.return_value = mock_service_catalog

        ctx = context.DesignateContext(
            auth_token='token', service_catalog='catalog'
        )

        auth_plugin = ctx.get_auth_plugin()
        auth_plugin.get_endpoint_data = mock.Mock()
        auth_plugin.get_endpoint(mock_session)

        mock_sc.assert_called_with('catalog')
        mock_service_catalog.url_for.assert_called_with(
            service_type=None, service_name=None, interface=None,
            region_name=None
        )
        auth_plugin.get_endpoint_data.assert_called()

    def test_get_auth_plugin_user(self):
        ctx = context.DesignateContext(
            user_auth_plugin='user_auth_plugin'
        )
        self.assertEqual('user_auth_plugin', ctx.get_auth_plugin())



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from oslo_config import fixture as cfg_fixture
import oslotest.base
import webtest

from designate.api import versions
from designate.common import constants
import designate.conf


CONF = designate.conf.CONF


class TestApiVersion(oslotest.base.BaseTestCase):

    def setUp(self):
        super().setUp()
        self.useFixture(cfg_fixture.Config(CONF))

    def test_add_a_version(self):
        api_url = 'http://203.0.113.1/v2'
        results = []

        versions._add_a_version(
            results, 'v2.1', api_url, constants.EXPERIMENTAL,
            '2022-08-10T00:00:00Z')

        self.assertEqual(1, len(results))
        self.assertEqual('v2.1', results[0]['id'])
        self.assertEqual(constants.EXPERIMENTAL, results[0]['status'])
        self.assertEqual('2022-08-10T00:00:00Z', results[0]['updated'])
        self.assertEqual(2, len(results[0]['links']))

    def test_get_versions(self):
        CONF.set_override('enable_host_header', False, 'service:api')
        CONF.set_override(
            'api_base_uri', 'http://203.0.113.1:9001/', 'service:api'
        )

        self.app = versions.factory({})
        self.client = webtest.TestApp(self.app)

        response = self.client.get('/')
        self.assertEqual(200, response.status_int)
        self.assertEqual('application/json', response.content_type)

        self.assertEqual(3, len(response.json['versions']))
        self.assertEqual(
            'http://203.0.113.1:9001/v2',
            response.json['versions'][0]['links'][0]['href']
        )

    def test_get_versions_with_enable_host_header(self):
        CONF.set_override('enable_host_header', True, 'service:api')

        self.app = versions.factory({})
        self.client = webtest.TestApp(self.app)

        response = self.client.get('/')
        self.assertEqual(200, response.status_int)
        self.assertEqual('application/json', response.content_type)

        self.assertEqual(3, len(response.json['versions']))
        self.assertEqual(
            'http://localhost/v2',
            response.json['versions'][0]['links'][0]['href']
        )



# Copyright 2015 Hewlett-Packard Development Company, L.P.
# Copyright 2022 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""domain_to_zone_rename

Revision ID: 867a331ce1fc
Revises: c9f427f7180a
Create Date: 2022-07-29 18:41:19.427853

"""
from alembic import op
import sqlalchemy as sa

from designate.storage.sqlalchemy.alembic import legacy_utils
from designate.storage.sqlalchemy.types import UUID

# revision identifiers, used by Alembic.
revision = '867a331ce1fc'
down_revision = 'c9f427f7180a'
branch_labels = None
depends_on = None

# Equivalent to legacy sqlalchemy-migrate revision 080_domain_to_zone_rename

# Notes from the original migration file:
# This migration removes all references to domain from our Database.
# We rename the domains and domain_attribute tables, and rename any columns
# that had "domain" in the name.as
# There is a follow on patch to recreate the FKs for the newly renamed
# tables as the lib we use doesn't seem to like creating FKs on renamed
# tables until after the migration is complete.


def _drop_foreign_key(fk_def):

    table = fk_def[0].table

    col = fk_def[0]
    ref_col = fk_def[1]

    # We need a naming convention to find unnamed foreign keys on sqlite
    naming_convention = {
        "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s"}

    # Use .copy() to avoid the set changing during the for operation
    # We need to search for the foreign keys as they may be named differently
    # between different dialects (mysql, sqlite, etc.)
    for fk in table.foreign_keys.copy():
        # We must use batch mode because the unit tests use sqlite
        with op.batch_alter_table(
                table.name, naming_convention=naming_convention) as batch_op:

            # Check if the fk is the one we want
            if fk.column == col and fk.parent == ref_col:
                batch_op.drop_constraint(fk.constraint.name,
                                         type_='foreignkey')

            # Check if the fk is the one we want (sometimes it seems the parent
            # / col is switched
            if fk.parent == col and fk.column == ref_col:
                batch_op.drop_constraint(fk.constraint.name,
                                         type_='foreignkey')


def upgrade() -> None:
    # Check if the equivalent legacy migration has already run
    if not legacy_utils.is_migration_needed(80):
        return

    convention = {
        "ix": 'ix_%(column_0_label)s',
        "uq": "uq_%(table_name)s_%(column_0_name)s",
        "ck": "ck_%(table_name)s_%(constraint_name)s",
        "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
        "pk": "pk_%(table_name)s"
    }

    metadata = sa.MetaData(naming_convention=convention)
    metadata.bind = op.get_bind()

    # Get all the tables
    domains_table = sa.Table('domains', metadata,
                             autoload_with=op.get_bind())
    domain_attrib_table = sa.Table('domain_attributes', metadata,
                                   autoload_with=op.get_bind())
    recordsets_table = sa.Table('recordsets', metadata,
                                autoload_with=op.get_bind())
    records_table = sa.Table('records', metadata, autoload_with=op.get_bind())
    ztr_table = sa.Table('zone_transfer_requests', metadata,
                         autoload_with=op.get_bind())
    zta_table = sa.Table('zone_transfer_accepts', metadata,
                         autoload_with=op.get_bind())

    # Remove the affected FKs
    # Define FKs
    fks = [
        [domains_table.c.id, domains_table.c.parent_domain_id],
        [domain_attrib_table.c.domain_id, domains_table.c.id],
        [recordsets_table.c.domain_id, domains_table.c.id],
        [records_table.c.domain_id, domains_table.c.id],
        [ztr_table.c.domain_id, domains_table.c.id],
        [zta_table.c.domain_id, domains_table.c.id]
    ]

    # Drop FKs
    for fk in fks:
        _drop_foreign_key(fk)

    with op.batch_alter_table('domains') as batch_op:
        batch_op.alter_column('parent_domain_id',
                              new_column_name='parent_zone_id',
                              existing_type=UUID)
    op.rename_table('domains', 'zones')

    with op.batch_alter_table('domain_attributes') as batch_op:
        batch_op.alter_column('domain_id', new_column_name='zone_id',
                              existing_type=UUID)
    op.rename_table('domain_attributes', 'zone_attributes')

    with op.batch_alter_table('recordsets') as batch_op:
        batch_op.alter_column('domain_id', new_column_name='zone_id',
                              existing_type=UUID)
        batch_op.alter_column('domain_shard', new_column_name='zone_shard',
                              existing_type=sa.SmallInteger)

    with op.batch_alter_table('records') as batch_op:
        batch_op.alter_column('domain_id', new_column_name='zone_id',
                              existing_type=UUID)
        batch_op.alter_column('domain_shard', new_column_name='zone_shard',
                              existing_type=sa.SmallInteger)

    with op.batch_alter_table('zone_transfer_requests') as batch_op:
        batch_op.alter_column('domain_id', new_column_name='zone_id',
                              existing_type=UUID)

    with op.batch_alter_table('zone_transfer_accepts') as batch_op:
        batch_op.alter_column('domain_id', new_column_name='zone_id',
                              existing_type=UUID)

    with op.batch_alter_table('zone_tasks') as batch_op:
        batch_op.alter_column('domain_id', new_column_name='zone_id',
                              existing_type=UUID)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from unittest import mock

import oslotest.base

from designate.api.v2.controllers import zones
from designate.central import rpcapi
from designate import objects


class TestZonesAPI(oslotest.base.BaseTestCase):
    def setUp(self):
        super().setUp()
        self.central_api = mock.Mock()
        self.zone = objects.Zone(
            id='1e8952a5-e5a4-426a-afab-4cd10131a351',
            name='example.com.',
            email='example@example.com',
            masters=objects.ZoneMasterList(),
            attributes=objects.ZoneAttributeList(),
        )
        mock.patch.object(rpcapi.CentralAPI, 'get_instance',
                          return_value=self.central_api).start()

        self.controller = zones.ZonesController()

    @mock.patch('pecan.response')
    @mock.patch('pecan.request')
    def test_post_all_zone_error(self, mock_request, mock_response):
        mock_response.headers = {}

        mock_request.environ = {'context': mock.Mock()}
        mock_request.body_dict = {
            'name': 'example.com.',
            'type': 'PRIMARY',
            'email': 'example@example.com',
        }

        zone = objects.Zone(
            name='example.com.',
            type='PRIMARY',
            email='example@example.com',
            status='ERROR',
            masters=objects.ZoneMasterList(),
            attributes=objects.ZoneAttributeList(),
        )

        self.central_api.create_zone.return_value = zone

        self.controller.post_all()

        self.assertEqual(201, mock_response.status_int)

    @mock.patch('pecan.response')
    @mock.patch('pecan.request')
    def test_patch_one_zone_error(self, mock_request, mock_response):
        mock_response.headers = {}

        mock_request.environ = {'context': mock.Mock()}
        mock_request.body_dict = {
            'name': 'example.com.',
            'type': 'PRIMARY',
            'email': 'example@example.com',
        }

        self.central_api.get_zone.return_value = self.zone
        self.central_api.update_zone.return_value = self.zone

        self.controller.patch_one(self.zone.id)

        self.assertEqual(200, mock_response.status_int)



# Copyright 2012 Managed I.T.
#
# Author: Kiall Mac Innes <kiall@managedit.ie>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from oslo_log import log as logging

import designate.conf
from designate.notification_handler.base import BaseAddressHandler


CONF = designate.conf.CONF
LOG = logging.getLogger(__name__)


class NovaFixedHandler(BaseAddressHandler):
    """Handler for Nova's notifications"""
    __plugin_name__ = 'nova_fixed'

    def get_exchange_topics(self):
        exchange = CONF[self.name].control_exchange

        topics = [topic for topic in CONF[self.name].notification_topics]

        return (exchange, topics)

    def get_event_types(self):
        return [
            'compute.instance.create.end',
            'compute.instance.delete.start',
        ]

    def _get_ip_data(self, addr_dict):
        data = super()._get_ip_data(addr_dict)
        data['label'] = addr_dict['label']
        return data

    def process_notification(self, context, event_type, payload):
        LOG.debug('NovaFixedHandler received notification - %s', event_type)

        zone_id = CONF[self.name].zone_id

        if not zone_id:
            LOG.error('NovaFixedHandler: zone_id is None, ignore the event.')
            return

        if event_type == 'compute.instance.create.end':
            payload['project'] = context.get('project_name', None)
            self._create(addresses=payload['fixed_ips'],
                         extra=payload,
                         zone_id=zone_id,
                         resource_id=payload['instance_id'],
                         resource_type='instance')

        elif event_type == 'compute.instance.delete.start':
            self._delete(zone_id=zone_id,
                         resource_id=payload['instance_id'],
                         resource_type='instance')



# Copyright 2013 Hewlett-Packard Development Company, L.P.
#
# Author: Kiall Mac Innes <kiall@hpe.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from oslo_log import log as logging
import pecan

from designate.api.v2.controllers import rest
from designate.api.v2.controllers.zones import nameservers
from designate.api.v2.controllers.zones import recordsets
from designate.api.v2.controllers.zones import sharedzones
from designate.api.v2.controllers.zones import tasks
import designate.conf
from designate import exceptions
from designate import objects
from designate.objects.adapters import DesignateAdapter
from designate import utils


CONF = designate.conf.CONF
LOG = logging.getLogger(__name__)


class ZonesController(rest.RestController):

    SORT_KEYS = ['created_at', 'id', 'updated_at', 'name', 'tenant_id',
                 'serial', 'ttl', 'status']

    recordsets = recordsets.RecordSetsController()
    tasks = tasks.TasksController()
    nameservers = nameservers.NameServersController()
    shares = sharedzones.SharedZonesController()

    @pecan.expose(template='json:', content_type='application/json')
    @utils.validate_uuid('zone_id')
    def get_one(self, zone_id):
        """Get Zone"""
        # TODO(kiall): Validate we have a sane UUID for zone_id

        request = pecan.request
        context = request.environ['context']

        zone = self.central_api.get_zone(context, zone_id)

        LOG.info('Retrieved %(zone)s', {'zone': zone})

        return DesignateAdapter.render('API_v2', zone, request=request)

    @pecan.expose(template='json:', content_type='application/json')
    def get_all(self, **params):
        """List Zones"""
        request = pecan.request
        context = request.environ['context']

        marker, limit, sort_key, sort_dir = utils.get_paging_params(
                context, params, self.SORT_KEYS)

        # Extract any filter params.
        accepted_filters = ('name', 'type', 'email', 'status',
                            'description', 'ttl', )

        criterion = self._apply_filter_params(
            params, accepted_filters, {})

        zones = self.central_api.find_zones(
            context, criterion, marker, limit, sort_key, sort_dir)

        LOG.info('Retrieved %(zones)s', {'zones': zones})

        return DesignateAdapter.render('API_v2', zones, request=request)

    @pecan.expose(template='json:', content_type='application/json')
    def post_all(self):
        """Create Zone"""
        request = pecan.request
        response = pecan.response
        context = request.environ['context']

        zone = request.body_dict

        if 'type' not in zone:
            zone['type'] = 'PRIMARY'

        zone = DesignateAdapter.parse('API_v2', zone, objects.Zone())
        zone.validate()

        if zone.type == 'SECONDARY':
            mgmt_email = CONF['service:central'].managed_resource_email
            zone['email'] = mgmt_email

        # Create the zone
        zone = self.central_api.create_zone(context, zone)

        # Shared is a virtual database column, so inject False here as a
        # new zone cannot yet be shared.
        zone.shared = False

        LOG.info('Created %(zone)s', {'zone': zone})

        # Prepare the response headers
        # If the zone has been created asynchronously
        if zone['status'] == 'PENDING':
            response.status_int = 202
        else:
            response.status_int = 201

        # Prepare and return the response body
        zone = DesignateAdapter.render('API_v2', zone, request=request)

        response.headers['Location'] = zone['links']['self']

        return zone

    @pecan.expose(template='json:', content_type='application/json')
    @pecan.expose(template='json:', content_type='application/json-patch+json')
    @utils.validate_uuid('zone_id')
    def patch_one(self, zone_id):
        """Update Zone"""
        # TODO(kiall): This needs cleanup to say the least..
        request = pecan.request
        context = request.environ['context']
        body = request.body_dict
        response = pecan.response

        # TODO(kiall): Validate we have a sane UUID for zone_id

        # Fetch the existing zone
        zone = self.central_api.get_zone(context, zone_id)

        # Don't allow updates to zones that are being deleted
        if zone.action == 'DELETE':
            raise exceptions.BadRequest('Can not update a deleting zone')

        if request.content_type == 'application/json-patch+json':
            # Possible pattern:
            #
            # 1) Load existing zone.
            # 2) Apply patch, maintain list of changes.
            # 3) Return changes, after passing through the code ^ for plain
            #    JSON.
            #
            # Difficulties:
            #
            # 1) "Nested" resources? records inside a recordset.
            # 2) What to do when a zone doesn't exist in the first place?
            # 3) ...?
            raise NotImplementedError('json-patch not implemented')
        else:
            # Update the zone object with the new values
            zone = DesignateAdapter.parse('API_v2', body, zone)

            zone.validate()
            # If masters are specified then we set zone.transferred_at to None
            # which will cause a new transfer
            if 'attributes' in zone.obj_what_changed():
                zone.transferred_at = None

            # Update and persist the resource

            increment_serial = zone.type == 'PRIMARY'
            zone = self.central_api.update_zone(
                context, zone, increment_serial=increment_serial)

        LOG.info('Updated %(zone)s', {'zone': zone})

        if zone.status == 'PENDING':
            response.status_int = 202
        else:
            response.status_int = 200

        return DesignateAdapter.render('API_v2', zone, request=request)

    @pecan.expose(template='json:', content_type='application/json')
    @utils.validate_uuid('zone_id')
    def delete_one(self, zone_id):
        """Delete Zone"""
        request = pecan.request
        response = pecan.response
        context = request.environ['context']

        zone = self.central_api.delete_zone(context, zone_id)
        response.status_int = 202

        LOG.info('Deleted %(zone)s', {'zone': zone})

        return DesignateAdapter.render('API_v2', zone, request=request)



# Copyright 2012 Managed I.T.
#
# Author: Kiall Mac Innes <kiall@managedit.ie>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import hashlib

from oslo_log import log as logging
from oslo_utils import timeutils
from sqlalchemy import case, select, distinct, func
from sqlalchemy.sql.expression import or_, literal_column

from designate.common import constants
from designate import exceptions
from designate import objects
from designate.objects.adapters import DesignateAdapter
from designate.storage import sql
from designate.storage.sqlalchemy import base
from designate.storage.sqlalchemy import tables

import designate.conf


CONF = designate.conf.CONF
LOG = logging.getLogger(__name__)

MAXIMUM_SUBZONE_DEPTH = 128


class SQLAlchemyStorage(base.SQLAlchemy):
    """SQLAlchemy connection"""
    __plugin_name__ = 'sqlalchemy'

    def __init__(self):
        super().__init__()

    def get_inspector(self):
        return sql.get_inspector()

    # CRUD for our resources (quota, server, tsigkey, tenant, zone & record)
    # R - get_*, find_*s
    #
    # Standard Arguments
    # self      - python object for the class
    # context   - a dictionary of details about the request (http etc),
    #             provided by flask.
    # criterion - dictionary of filters to be applied
    #

    # Quota Methods
    def _find_quotas(self, context, criterion, one=False, marker=None,
                     limit=None, sort_key=None, sort_dir=None):
        return self._find(
            context, tables.quotas, objects.Quota, objects.QuotaList,
            exceptions.QuotaNotFound, criterion, one, marker, limit,
            sort_key, sort_dir)

    def create_quota(self, context, quota):
        """
        Create a Quota.

        :param context: RPC Context.
        :param quota: Quota object with the values to be created.
        """
        if not isinstance(quota, objects.Quota):
            # TODO(kiall): Quotas should always use Objects
            quota = objects.Quota(**quota)

        return self._create(
            tables.quotas, quota, exceptions.DuplicateQuota)

    def get_quota(self, context, quota_id):
        """
        Get a Quota via ID.

        :param context: RPC Context.
        :param quota_id: Quota ID to get.
        """
        return self._find_quotas(context, {'id': quota_id}, one=True)

    def find_quotas(self, context, criterion=None, marker=None, limit=None,
                    sort_key=None, sort_dir=None):
        """
        Find Quotas

        :param context: RPC Context.
        :param criterion: Criteria to filter by.
        :param marker: Resource ID from which after the requested page will
                       start after
        :param limit: Integer limit of objects of the page size after the
                      marker
        :param sort_key: Key from which to sort after.
        :param sort_dir: Direction to sort after using sort_key.
        """
        return self._find_quotas(context, criterion, marker=marker,
                                 limit=limit, sort_key=sort_key,
                                 sort_dir=sort_dir)

    def find_quota(self, context, criterion):
        """
        Find a single Quota.

        :param context: RPC Context.
        :param criterion: Criteria to filter by.
        """
        return self._find_quotas(context, criterion, one=True)

    def update_quota(self, context, quota):
        """
        Update a Quota

        :param context: RPC Context.
        :param quota: Quota to update.
        """
        return self._update(
            context, tables.quotas, quota, exceptions.DuplicateQuota,
            exceptions.QuotaNotFound)

    def delete_quota(self, context, quota_id):
        """
        Delete a Quota via ID.

        :param context: RPC Context.
        :param quota_id: Delete a Quota via ID
        """
        # Fetch the existing quota, we'll need to return it.
        quota = self._find_quotas(context, {'id': quota_id}, one=True)
        return self._delete(context, tables.quotas, quota,
                            exceptions.QuotaNotFound)

    # TLD Methods
    def _find_tlds(self, context, criterion, one=False, marker=None,
                   limit=None, sort_key=None, sort_dir=None):
        return self._find(
            context, tables.tlds, objects.Tld, objects.TldList,
            exceptions.TldNotFound, criterion, one, marker, limit,
            sort_key, sort_dir)

    def create_tld(self, context, tld):
        """
        Create a TLD.

        :param context: RPC Context.
        :param tld: Tld object with the values to be created.
        """
        return self._create(
            tables.tlds, tld, exceptions.DuplicateTld)

    def get_tld(self, context, tld_id):
        """
        Get a TLD via ID.

        :param context: RPC Context.
        :param tld_id: TLD ID to get.
        """
        return self._find_tlds(context, {'id': tld_id}, one=True)

    def find_tlds(self, context, criterion=None, marker=None, limit=None,
                  sort_key=None, sort_dir=None):
        """
        Find TLDs

        :param context: RPC Context.
        :param criterion: Criteria to filter by.
        :param marker: Resource ID from which after the requested page will
                       start after
        :param limit: Integer limit of objects of the page size after the
                      marker
        :param sort_key: Key from which to sort after.
        :param sort_dir: Direction to sort after using sort_key.
        """
        return self._find_tlds(context, criterion, marker=marker, limit=limit,
                               sort_key=sort_key, sort_dir=sort_dir)

    def find_tld(self, context, criterion):
        """
        Find a single TLD.

        :param context: RPC Context.
        :param criterion: Criteria to filter by.
        """
        return self._find_tlds(context, criterion, one=True)

    def update_tld(self, context, tld):
        """
        Update a TLD

        :param context: RPC Context.
        :param tld: TLD to update.
        """
        return self._update(
            context, tables.tlds, tld, exceptions.DuplicateTld,
            exceptions.TldNotFound)

    def delete_tld(self, context, tld_id):
        """
        Delete a TLD via ID.

        :param context: RPC Context.
        :param tld_id: Delete a TLD via ID
        """
        # Fetch the existing tld, we'll need to return it.
        tld = self._find_tlds(context, {'id': tld_id}, one=True)
        return self._delete(context, tables.tlds, tld, exceptions.TldNotFound)

    # TSIG Key Methods
    def _find_tsigkeys(self, context, criterion, one=False, marker=None,
                       limit=None, sort_key=None, sort_dir=None):
        return self._find(
            context, tables.tsigkeys, objects.TsigKey, objects.TsigKeyList,
            exceptions.TsigKeyNotFound, criterion, one, marker, limit,
            sort_key, sort_dir)

    def create_tsigkey(self, context, tsigkey):
        """
        Create a TSIG Key.

        :param context: RPC Context.
        :param tsigkey: TsigKey object with the values to be created.
        """
        return self._create(
            tables.tsigkeys, tsigkey, exceptions.DuplicateTsigKey)

    def get_tsigkey(self, context, tsigkey_id):
        """
        Get a TSIG Key via ID.

        :param context: RPC Context.
        :param tsigkey_id: Server ID to get.
        """
        return self._find_tsigkeys(context, {'id': tsigkey_id}, one=True)

    def find_tsigkeys(self, context, criterion=None, marker=None, limit=None,
                      sort_key=None, sort_dir=None):
        """
        Find TSIG Keys.

        :param context: RPC Context.
        :param criterion: Criteria to filter by.
        :param marker: Resource ID from which after the requested page will
                       start after
        :param limit: Integer limit of objects of the page size after the
                      marker
        :param sort_key: Key from which to sort after.
        :param sort_dir: Direction to sort after using sort_key.
        """
        return self._find_tsigkeys(context, criterion, marker=marker,
                                   limit=limit, sort_key=sort_key,
                                   sort_dir=sort_dir)

    def find_tsigkey(self, context, criterion):
        """
        Find TSIG Key.

        :param context: RPC Context.
        :param criterion: Criteria to filter by.
        """
        return self._find_tsigkeys(context, criterion, one=True)

    def update_tsigkey(self, context, tsigkey):
        """
        Update a TSIG Key

        :param context: RPC Context.
        :param tsigkey: TSIG Keyto update.
        """
        return self._update(
            context, tables.tsigkeys, tsigkey, exceptions.DuplicateTsigKey,
            exceptions.TsigKeyNotFound)

    def delete_tsigkey(self, context, tsigkey_id):
        """
        Delete a TSIG Key via ID.

        :param context: RPC Context.
        :param tsigkey_id: Delete a TSIG Key via ID
        """
        # Fetch the existing tsigkey, we'll need to return it.
        tsigkey = self._find_tsigkeys(context, {'id': tsigkey_id}, one=True)
        return self._delete(context, tables.tsigkeys, tsigkey,
                            exceptions.TsigKeyNotFound)

    ##
    # Tenant Methods
    ##
    def find_tenants(self, context):
        """
        Find all Tenants.

        :param context: RPC Context.
        """
        # returns an array of tenant_id & count of their zones
        query = select(tables.zones.c.tenant_id, func.count(tables.zones.c.id))
        query = self._apply_tenant_criteria(context, tables.zones, query)
        query = self._apply_deleted_criteria(context, tables.zones, query)
        query = query.group_by(tables.zones.c.tenant_id)

        with sql.get_read_session() as session:
            resultproxy = session.execute(query)
            results = resultproxy.fetchall()

        tenant_list = objects.TenantList(
            objects=[objects.Tenant(id=t[0], zone_count=t[1]) for t in
                     results])

        tenant_list.obj_reset_changes()

        return tenant_list

    def get_tenant(self, context, tenant_id):
        """
        Get Tenant.

        :param context: RPC Context.
        :param tenant_id: ID of the Tenant.
        """
        # get list & count of all zones owned by given tenant_id
        query = select(tables.zones.c.name)
        query = self._apply_tenant_criteria(context, tables.zones, query)
        query = self._apply_deleted_criteria(context, tables.zones, query)
        query = query.where(tables.zones.c.tenant_id == tenant_id)

        with sql.get_read_session() as session:
            resultproxy = session.execute(query)
            results = resultproxy.fetchall()

        return objects.Tenant(
            id=tenant_id,
            zone_count=len(results),
            zones=[r[0] for r in results])

    def count_tenants(self, context):
        """
        Count tenants

        :param context: RPC Context.
        """
        # tenants are the owner of zones, count the number of unique tenants
        # select count(distinct tenant_id) from zones
        query = select(func.count(distinct(tables.zones.c.tenant_id)))
        query = self._apply_tenant_criteria(context, tables.zones, query)
        query = self._apply_deleted_criteria(context, tables.zones, query)

        with sql.get_read_session() as session:
            resultproxy = session.execute(query)
            result = resultproxy.fetchone()

        if result is None:
            return 0

        return result[0]

    ##
    # Zone Methods
    ##
    def _find_zones(self, context, criterion, one=False, marker=None,
                    limit=None, sort_key=None, sort_dir=None,
                    apply_tenant_criteria=True, include_shared=False):
        # Check to see if the criterion can use the reverse_name column
        criterion = self._rname_check(criterion)

        # Create a virtual column showing if the zone is shared or not.
        shared_case = case((tables.shared_zones.c.target_project_id.is_(None),
                            literal_column('False')),
                           else_=literal_column('True')).label('shared')
        query = select(
            tables.zones,
            shared_case).outerjoin(tables.shared_zones).distinct()

        zones = self._find(
            context, tables.zones, objects.Zone, objects.ZoneList,
            exceptions.ZoneNotFound, criterion, one, marker, limit,
            sort_key, sort_dir, query=query,
            apply_tenant_criteria=apply_tenant_criteria,
            include_shared=include_shared)

        def _load_relations(zone):
            if zone.type == 'SECONDARY':
                zone.masters = self._find_zone_masters(
                    context, {'zone_id': zone.id})
            else:
                # This avoids an extra DB call per primary zone. This will
                # always have 0 results for a PRIMARY zone.
                zone.masters = objects.ZoneMasterList()

            zone.attributes = self._find_zone_attributes(
                context, {'zone_id': zone.id, 'key': '!master'})

            zone.obj_reset_changes(['masters', 'attributes'])

        # TODO(Federico) refactor part of _find_zones into _find_zone, move
        # _load_relations out

        if one:
            _load_relations(zones)
        else:
            zones.total_count = self.count_zones(context, criterion)
            for d in zones:
                _load_relations(d)

        if one:
            LOG.debug('Fetched zone %s', zones)
        return zones

    def create_zone(self, context, zone):
        """
        Create a new Zone.

        :param context: RPC Context.
        :param zone: Zone object with the values to be created.
        """
        # Patch in the reverse_name column
        extra_values = {'reverse_name': zone.name[::-1]}

        # Don't handle recordsets for now
        zone = self._create(
            tables.zones, zone, exceptions.DuplicateZone,
            ['attributes', 'recordsets', 'masters'],
            extra_values=extra_values)

        if zone.obj_attr_is_set('attributes'):
            for attrib in zone.attributes:
                self.create_zone_attribute(context, zone.id, attrib)
        else:
            zone.attributes = objects.ZoneAttributeList()
        if zone.obj_attr_is_set('masters'):
            for master in zone.masters:
                self.create_zone_master(context, zone.id, master)
        else:
            zone.masters = objects.ZoneMasterList()
        zone.obj_reset_changes(['masters', 'attributes'])

        return zone

    def get_zone(self, context, zone_id, apply_tenant_criteria=True):
        """
        Get a Zone via its ID.

        :param context: RPC Context.
        :param zone_id: ID of the Zone.
        :param apply_tenant_criteria: Whether to filter results by project_id.
        """
        zone = self._find_zones(context, {'id': zone_id}, one=True,
                                apply_tenant_criteria=apply_tenant_criteria,
                                include_shared=True)
        return zone

    def find_zones(self, context, criterion=None, marker=None, limit=None,
                   sort_key=None, sort_dir=None):
        """
        Find zones

        :param context: RPC Context.
        :param criterion: Criteria to filter by.
        :param marker: Resource ID from which after the requested page will
                       start after
        :param limit: Integer limit of objects of the page size after the
                      marker
        :param sort_key: Key from which to sort after.
        :param sort_dir: Direction to sort after using sort_key.
        """
        zones = self._find_zones(context, criterion, marker=marker,
                                 limit=limit, sort_key=sort_key,
                                 sort_dir=sort_dir, include_shared=True)
        return zones

    def find_zone(self, context, criterion):
        """
        Find a single Zone.

        :param context: RPC Context.
        :param criterion: Criteria to filter by.
        """
        zone = self._find_zones(context, criterion, one=True,
                                include_shared=True)
        return zone

    def update_zone(self, context, zone):
        """
        Update a Zone

        :param context: RPC Context.
        :param zone: Zone object.
        """
        tenant_id_changed = False
        if 'tenant_id' in zone.obj_what_changed():
            tenant_id_changed = True

        # Don't handle recordsets for now
        LOG.debug('Updating zone %s', zone)
        updated_zone = self._update(
            context, tables.zones, zone, exceptions.DuplicateZone,
            exceptions.ZoneNotFound,
            ['attributes', 'recordsets', 'masters'])

        if zone.obj_attr_is_set('attributes'):
            # Gather the Attribute ID's we have
            have = {r.id for r in self._find_zone_attributes(
                context, {'zone_id': zone.id})}

            # Prep some lists of changes
            keep = set()
            create = []
            update = []

            # Determine what to change
            for i in zone.attributes:
                keep.add(i.id)
                try:
                    i.obj_get_original_value('id')
                except KeyError:
                    create.append(i)
                else:
                    update.append(i)

            # NOTE: Since we're dealing with mutable objects, the return value
            #       of create/update/delete attribute is not needed.
            #       The original item will be mutated in place on the input
            #       "zone.attributes" list.

            # Delete Attributes
            for i_id in have - keep:
                attr = self._find_zone_attributes(
                    context, {'id': i_id}, one=True)
                self.delete_zone_attribute(context, attr.id)

            # Update Attributes
            for i in update:
                self.update_zone_attribute(context, i)

            # Create Attributes
            for attr in create:
                attr.zone_id = zone.id
                self.create_zone_attribute(context, zone.id, attr)

        if zone.obj_attr_is_set('masters'):
            # Gather the Attribute ID's we have
            have = {r.id for r in self._find_zone_masters(
                context, {'zone_id': zone.id})}

            # Prep some lists of changes
            keep = set()
            create = []
            update = []

            # Determine what to change
            for i in zone.masters:
                keep.add(i.id)
                try:
                    i.obj_get_original_value('id')
                except KeyError:
                    create.append(i)
                else:
                    update.append(i)

            # NOTE: Since we're dealing with mutable objects, the return value
            #       of create/update/delete attribute is not needed.
            #       The original item will be mutated in place on the input
            #       "zone.attributes" list.

            # Delete Attributes
            for i_id in have - keep:
                attr = self._find_zone_masters(
                    context, {'id': i_id}, one=True)
                self.delete_zone_master(context, attr.id)

            # Update Attributes
            for i in update:
                self.update_zone_master(context, i)

            # Create Attributes
            for attr in create:
                attr.zone_id = zone.id
                self.create_zone_master(context, zone.id, attr)

        if zone.obj_attr_is_set('recordsets'):
            existing = {}
            for rrset in self.find_recordsets(context, {'zone_id': zone.id}):
                existing[rrset.name, rrset.type] = rrset

            keep = set()
            for rrset in zone.recordsets:
                existing_recordset = existing.get((rrset.name, rrset.type))
                if existing_recordset:
                    existing_recordset.update(rrset)
                    existing_recordset.records = rrset.records
                    existing_recordset.obj_reset_changes(['zone_name'])
                    self.update_recordset(context, existing_recordset)
                    keep.add(existing_recordset.id)
                else:
                    self.create_recordset(context, zone.id, rrset)
                    keep.add(rrset.id)

            if zone.type == 'SECONDARY':
                # Purge anything that shouldn't be there :P
                for i in {i.id for i in existing.values()} - keep:
                    self.delete_recordset(context, i)

        if tenant_id_changed:
            with sql.get_write_session() as session:
                session.execute(
                    tables.recordsets.update().
                    where(tables.recordsets.c.zone_id == zone.id).
                    values({'tenant_id': zone.tenant_id})
                )

                session.execute(
                    tables.records.update().
                    where(tables.records.c.zone_id == zone.id).
                    values({'tenant_id': zone.tenant_id})
                )

        return updated_zone

    def increment_serial(self, context, zone_id):
        """Increment the zone's serial number.
        """
        new_serial = timeutils.utcnow_ts()
        query = tables.zones.update().where(
            tables.zones.c.id == zone_id).values(
            {'serial': new_serial, 'increment_serial': False}
        )
        with sql.get_write_session() as session:
            session.execute(query)
        LOG.debug('Incremented zone serial for %s to %d', zone_id, new_serial)
        return new_serial

    def delete_zone(self, context, zone_id):
        """
        Delete a Zone

        :param context: RPC Context.
        :param zone_id: Zone ID to delete.
        """
        # Fetch the existing zone, we'll need to return it.
        zone = self._find_zones(context, {'id': zone_id}, one=True)
        return self._delete(context, tables.zones, zone,
                            exceptions.ZoneNotFound)

    def purge_zone(self, context, zone):
        """
        Purge a Zone, effectively removing the zone database record.

        :param context: RPC Context.
        :param zone: Zone to delete.
        """
        return self._delete(context, tables.zones, zone,
                            exceptions.ZoneNotFound, hard_delete=True)

    def _walk_up_zones(self, current, zones_by_id):
        """Walk upwards in a zone hierarchy until we find a parent zone
        that does not belong to "zones_by_id"
        :returns: parent zone ID or None
        """
        max_steps = MAXIMUM_SUBZONE_DEPTH
        while current.parent_zone_id in zones_by_id:
            current = zones_by_id[current.parent_zone_id]
            max_steps -= 1
            if max_steps == 0:
                raise exceptions.IllegalParentZone('Loop detected in the'
                                                   ' zone hierarchy')

        return current.parent_zone_id

    def purge_zones(self, context, criterion, limit):
        """
        Purge Zones, effectively removing the zones database records.

        Reparent orphan childrens, if any.
        Transactions/locks are not needed.

        :param context: RPC Context.
        :param criterion: Criteria to filter by.
        :param limit: Integer limit of objects of the page size after the
                      marker
        """
        if 'deleted' in criterion:
            context.show_deleted = True

        zones = self.find_zones(
            context=context,
            criterion=criterion,
            limit=limit,
        )
        if not zones:
            LOG.info('No zones to be purged')
            return

        LOG.debug('Purging %d zones', len(zones))

        zones_by_id = {z.id: z for z in zones}

        for zone in zones:

            # Reparent child zones, if any.
            surviving_parent_id = self._walk_up_zones(zone, zones_by_id)
            query = (
                tables.zones.update().
                where(tables.zones.c.parent_zone_id == zone.id).
                values(parent_zone_id=surviving_parent_id)
            )

            with sql.get_write_session() as session:
                resultproxy = session.execute(query)
                LOG.debug('%d child zones updated', resultproxy.rowcount)

            self.purge_zone(context, zone)

        LOG.info('Purged %d zones', len(zones))
        return len(zones)

    def count_zones(self, context, criterion=None):
        """
        Count zones

        :param context: RPC Context.
        :param criterion: Criteria to filter by.
        """
        query = select(func.count(tables.zones.c.id))
        query = self._apply_criterion(tables.zones, query, criterion)
        query = self._apply_tenant_criteria(context, tables.zones, query)
        query = self._apply_deleted_criteria(context, tables.zones, query)

        with sql.get_read_session() as session:
            resultproxy = session.execute(query)
            result = resultproxy.fetchone()

        if result is None:
            return 0

        return result[0]

    # Shared zones methods
    def _find_shared_zones(self, context, criterion, one=False, marker=None,
                           limit=None, sort_key=None, sort_dir=None):

        table = tables.shared_zones

        query = select(table)

        if not context.all_tenants:
            query = query.where(or_(
                table.c.project_id == context.project_id,
                table.c.target_project_id == context.project_id))

        return self._find(
            context, tables.shared_zones, objects.SharedZone,
            objects.SharedZoneList, exceptions.SharedZoneNotFound, criterion,
            one, marker, limit, sort_key, sort_dir, query=query,
            apply_tenant_criteria=False)

    def share_zone(self, context, shared_zone):
        """
        Share zone

        :param context: RPC Context.
        :param shared_zone: Shared Zone dict
        """
        return self._create(tables.shared_zones, shared_zone,
                            exceptions.DuplicateSharedZone)

    def unshare_zone(self, context, zone_id, shared_zone_id):
        """
        Unshare zone

        :param context: RPC Context.
        :param shared_zone_id: Shared Zone Id
        """
        shared_zone = self._find_shared_zones(
            context, {'id': shared_zone_id, 'zone_id': zone_id}, one=True
        )
        return self._delete(context, tables.shared_zones, shared_zone,
                            exceptions.SharedZoneNotFound)

    def find_shared_zones(self, context, criterion=None, marker=None,
                          limit=None, sort_key=None, sort_dir=None):
        """
        Find shared zones

        :param context: RPC Context.
        :param criterion: Criteria to filter by.
        :param marker: Resource ID from which after the requested page will
                       start after
        :param limit: Integer limit of objects of the page size after the
                      marker
        :param sort_key: Key from which to sort after.
        :param sort_dir: Direction to sort after using sort_key.
        """
        return self._find_shared_zones(
            context, criterion, marker=marker,
            limit=limit, sort_key=sort_key, sort_dir=sort_dir
        )

    def get_shared_zone(self, context, zone_id, shared_zone_id):
        """
        Get a shared zone via ID

        :param context: RPC Context.
        :param shared_zone_id: Shared Zone Id
        """
        return self._find_shared_zones(
            context, {'id': shared_zone_id, 'zone_id': zone_id}, one=True
        )

    def is_zone_shared_with_project(self, zone_id, project_id):
        """
        Checks if a zone is shared with a project.

        :param zone_id: The zone ID to check.
        :param project_id: The project ID to check.
        :returns: Boolean True/False if the zone is shared with the project.
        """
        query = select(literal_column('true'))
        query = query.where(tables.shared_zones.c.zone_id == zone_id)
        query = query.where(
            tables.shared_zones.c.target_project_id == project_id)
        with sql.get_read_session() as session:
            return session.scalar(query) is not None

    def delete_zone_shares(self, zone_id):
        """
        Delete all of the zone shares for a specific zone.

        :param zone_id: The zone ID to check.
        """
        query = tables.shared_zones.delete().where(
            tables.shared_zones.c.zone_id == zone_id)
        with sql.get_write_session() as session:
            session.execute(query)

    # Zone attribute methods
    def _find_zone_attributes(self, context, criterion, one=False,
                              marker=None, limit=None, sort_key=None,
                              sort_dir=None):
        return self._find(context, tables.zone_attributes,
                          objects.ZoneAttribute, objects.ZoneAttributeList,
                          exceptions.ZoneAttributeNotFound, criterion, one,
                          marker, limit, sort_key, sort_dir)

    def create_zone_attribute(self, context, zone_id, zone_attribute):
        zone_attribute.zone_id = zone_id
        return self._create(tables.zone_attributes, zone_attribute,
                            exceptions.DuplicateZoneAttribute)

    def get_zone_attributes(self, context, zone_attribute_id):
        return self._find_zone_attributes(
            context, {'id': zone_attribute_id}, one=True)

    def find_zone_attributes(self, context, criterion=None, marker=None,
                             limit=None, sort_key=None, sort_dir=None):
        return self._find_zone_attributes(context, criterion, marker=marker,
                                          limit=limit, sort_key=sort_key,
                                          sort_dir=sort_dir)

    def update_zone_attribute(self, context, zone_attribute):
        return self._update(context, tables.zone_attributes,
                            zone_attribute,
                            exceptions.DuplicateZoneAttribute,
                            exceptions.ZoneAttributeNotFound)

    def delete_zone_attribute(self, context, zone_attribute_id):
        zone_attribute = self._find_zone_attributes(
            context, {'id': zone_attribute_id}, one=True)
        deleted_zone_attribute = self._delete(
            context, tables.zone_attributes, zone_attribute,
            exceptions.ZoneAttributeNotFound)

        return deleted_zone_attribute

    # Zone master methods
    def _find_zone_masters(self, context, criterion, one=False,
                           marker=None, limit=None, sort_key=None,
                           sort_dir=None):

        return self._find(context, tables.zone_masters,
                          objects.ZoneMaster, objects.ZoneMasterList,
                          exceptions.ZoneMasterNotFound, criterion, one,
                          marker, limit, sort_key, sort_dir)

    def create_zone_master(self, context, zone_id, zone_master):
        zone_master.zone_id = zone_id
        return self._create(tables.zone_masters, zone_master,
                            exceptions.DuplicateZoneMaster)

    def update_zone_master(self, context, zone_master):
        return self._update(context, tables.zone_masters,
                            zone_master,
                            exceptions.DuplicateZoneMaster,
                            exceptions.ZoneMasterNotFound)

    def delete_zone_master(self, context, zone_master_id):
        zone_master = self._find_zone_masters(
            context, {'id': zone_master_id}, one=True)
        deleted_zone_master = self._delete(
            context, tables.zone_masters, zone_master,
            exceptions.ZoneMasterNotFound)

        return deleted_zone_master

    # RecordSet Methods
    def _find_recordsets(self, context, criterion, one=False, marker=None,
                         limit=None, sort_key=None, sort_dir=None,
                         force_index=False, apply_tenant_criteria=True):

        # Check to see if the criterion can use the reverse_name column
        criterion = self._rname_check(criterion)

        if criterion is not None and not criterion.get('zones_deleted', True):
            # remove 'zones_deleted' from the criterion, as _apply_criterion
            # assumes each key in criterion to be a column name.
            del criterion['zones_deleted']

        if one:
            rjoin = tables.recordsets.join(
                tables.zones,
                tables.recordsets.c.zone_id == tables.zones.c.id)
            query = (
                select(tables.recordsets).select_from(rjoin).
                where(tables.zones.c.deleted == '0')
            )

            recordsets = self._find(
                context, tables.recordsets, objects.RecordSet,
                objects.RecordSetList, exceptions.RecordSetNotFound, criterion,
                one, marker, limit, sort_key, sort_dir, query,
                apply_tenant_criteria=apply_tenant_criteria,
            )

            recordsets.records = self._find_records(
                context, {'recordset_id': recordsets.id},
                apply_tenant_criteria=apply_tenant_criteria,
            )

            recordsets.obj_reset_changes(['records'])

        else:
            tc, recordsets = self._find_recordsets_with_records(
                    context, criterion, limit=limit, marker=marker,
                    sort_key=sort_key, sort_dir=sort_dir,
                    force_index=force_index,
                    apply_tenant_criteria=apply_tenant_criteria,
            )

            recordsets.total_count = tc

        return recordsets

    def find_recordsets_axfr(self, context, criterion=None):
        """
        Find RecordSets.

        :param context: RPC Context.
        :param criterion: Criteria to filter by.
        """
        # Check to see if the criterion can use the reverse_name column
        criterion = self._rname_check(criterion)

        rjoin = tables.records.join(
            tables.recordsets,
            tables.records.c.recordset_id == tables.recordsets.c.id)

        query = (
            select(tables.recordsets.c.id, tables.recordsets.c.type,
                   tables.recordsets.c.ttl, tables.recordsets.c.name,
                   tables.records.c.data, tables.records.c.action).
            select_from(rjoin).where(tables.records.c.action != 'DELETE')
        )

        query = query.order_by(tables.recordsets.c.id)

        raw_rows = self._select_raw(
            context, tables.recordsets, criterion, query)

        return raw_rows

    def create_recordset(self, context, zone_id, recordset):
        """
        Create a recordset on a given Zone ID

        :param context: RPC Context.
        :param zone_id: Zone ID to create the recordset in.
        :param recordset: RecordSet object with the values to be created.
        """
        recordset.tenant_id = context.project_id
        recordset.zone_id = zone_id

        # Patch in the reverse_name column
        extra_values = {'reverse_name': recordset.name[::-1]}

        recordset = self._create(
            tables.recordsets, recordset, exceptions.DuplicateRecordSet,
            ['records'], extra_values=extra_values)

        if recordset.obj_attr_is_set('records'):
            for record in recordset.records:
                # NOTE: Since we're dealing with a mutable object, the return
                #       value is not needed. The original item will be mutated
                #       in place on the input "recordset.records" list.
                self.create_record(context, zone_id, recordset.id, record)
        else:
            recordset.records = objects.RecordList()

        recordset.obj_reset_changes(['records'])

        return recordset

    def find_recordsets_export(self, context, criterion=None):
        query = None

        rjoin = tables.records.join(
            tables.recordsets,
            tables.records.c.recordset_id == tables.recordsets.c.id)

        query = (
            select(tables.recordsets.c.name, tables.recordsets.c.ttl,
                   tables.recordsets.c.type, tables.records.c.data).
            select_from(rjoin)
        )

        query = query.order_by(tables.recordsets.c.created_at)

        raw_rows = self._select_raw(
            context, tables.recordsets, criterion, query)

        return raw_rows

    def find_recordsets(self, context, criterion=None, marker=None, limit=None,
                        sort_key=None, sort_dir=None, force_index=False,
                        apply_tenant_criteria=True):
        """
        Find RecordSets.

        :param context: RPC Context.
        :param criterion: Criteria to filter by.
        :param marker: Resource ID from which after the requested page will
                       start after
        :param limit: Integer limit of objects of the page size after the
                      marker
        :param sort_key: Key from which to sort after.
        :param sort_dir: Direction to sort after using sort_key.
        :param apply_tenant_criteria: Whether to filter results by project_id.
        """
        return self._find_recordsets(
            context, criterion, marker=marker, sort_dir=sort_dir,
            sort_key=sort_key, limit=limit, force_index=force_index,
            apply_tenant_criteria=apply_tenant_criteria)

    def find_recordset(self, context, criterion, apply_tenant_criteria=True):
        """
        Find a single RecordSet.

        :param context: RPC Context.
        :param criterion: Criteria to filter by.
        :param apply_tenant_criteria: Whether to filter results by project_id.
        """
        return self._find_recordsets(
            context, criterion, one=True,
            apply_tenant_criteria=apply_tenant_criteria)

    def update_recordset(self, context, recordset):
        """
        Update a recordset

        :param context: RPC Context.
        :param recordset: RecordSet to update
        """
        recordset = self._update(
            context, tables.recordsets, recordset,
            exceptions.DuplicateRecordSet, exceptions.RecordSetNotFound,
            ['records'])

        if recordset.obj_attr_is_set('records'):
            # Gather the Record ID's we have
            have_records = {r.id for r in self._find_records(
                context, {'recordset_id': recordset.id})}

            # Prep some lists of changes
            keep_records = set()
            create_records = []
            update_records = []

            # Determine what to change
            for record in recordset.records:
                keep_records.add(record.id)
                try:
                    record.obj_get_original_value('id')
                except KeyError:
                    create_records.append(record)
                else:
                    update_records.append(record)

            # NOTE: Since we're dealing with mutable objects, the return value
            #       of create/update/delete record is not needed. The original
            #       item will be mutated in place on the input
            #       "recordset.records" list.

            # Delete Records
            for record_id in have_records - keep_records:
                self.delete_record(context, record_id)

            # Update Records
            for record in update_records:
                self.update_record(context, record)

            # Create Records
            for record in create_records:
                self.create_record(
                    context, recordset.zone_id, recordset.id, record)

        return recordset

    def delete_recordset(self, context, recordset_id):
        """
        Delete a recordset

        :param context: RPC Context.
        :param recordset_id: RecordSet ID to delete
        """
        # Fetch the existing recordset, we'll need to return it.
        recordset = self._find_recordsets(
            context, {'id': recordset_id}, one=True)

        return self._delete(context, tables.recordsets, recordset,
                            exceptions.RecordSetNotFound)

    def count_recordsets(self, context, criterion=None):
        """
        Count recordsets

        :param context: RPC Context.
        :param criterion: Criteria to filter by.
        """
        # Ensure that we return only active recordsets
        rjoin = tables.recordsets.join(
            tables.zones,
            tables.recordsets.c.zone_id == tables.zones.c.id)

        query = (
            select(func.count(tables.recordsets.c.id)).
            select_from(rjoin).
            where(tables.zones.c.deleted == '0')
        )

        query = self._apply_criterion(tables.recordsets, query, criterion)
        query = self._apply_tenant_criteria(context, tables.recordsets, query)
        query = self._apply_deleted_criteria(context, tables.recordsets, query)

        with sql.get_read_session() as session:
            resultproxy = session.execute(query)
            result = resultproxy.fetchone()

        if result is None:
            return 0

        return result[0]

    # Record Methods
    def _find_records(self, context, criterion, one=False, marker=None,
                      limit=None, sort_key=None, sort_dir=None,
                      apply_tenant_criteria=True):
        return self._find(
            context, tables.records, objects.Record, objects.RecordList,
            exceptions.RecordNotFound, criterion, one, marker, limit,
            sort_key, sort_dir,
            apply_tenant_criteria=apply_tenant_criteria,
        )

    def _recalculate_record_hash(self, record):
        """
        Calculates the hash of the record, used to ensure record uniqueness.
        """
        md5sum = hashlib.md5(usedforsecurity=False)
        md5sum.update(('{}:{}'.format(record.recordset_id,
                                      record.data)).encode('utf-8'))

        return md5sum.hexdigest()

    def create_record(self, context, zone_id, recordset_id, record):
        """
        Create a record on a given Zone ID

        :param context: RPC Context.
        :param zone_id: Zone ID to create the record in.
        :param recordset_id: RecordSet ID to create the record in.
        :param record: Record object with the values to be created.
        """
        record.tenant_id = context.project_id
        record.zone_id = zone_id
        record.recordset_id = recordset_id
        record.hash = self._recalculate_record_hash(record)

        return self._create(
            tables.records, record, exceptions.DuplicateRecord)

    def get_record(self, context, record_id):
        """
        Get a record via ID

        :param context: RPC Context.
        :param record_id: Record ID to get
        """
        return self._find_records(context, {'id': record_id}, one=True)

    def find_records(self, context, criterion=None, marker=None, limit=None,
                     sort_key=None, sort_dir=None):
        """
        Find Records.

        :param context: RPC Context.
        :param criterion: Criteria to filter by.
        :param marker: Resource ID from which after the requested page will
                       start after
        :param limit: Integer limit of objects of the page size after the
                      marker
        :param sort_key: Key from which to sort after.
        :param sort_dir: Direction to sort after using sort_key.
        """
        return self._find_records(context, criterion, marker=marker,
                                  limit=limit, sort_key=sort_key,
                                  sort_dir=sort_dir)

    def find_record(self, context, criterion):
        """
        Find a single Record.

        :param context: RPC Context.
        :param criterion: Criteria to filter by.
        """
        return self._find_records(context, criterion, one=True)

    def update_record(self, context, record):
        """
        Update a record

        :param context: RPC Context.
        :param record: Record to update
        """
        if record.obj_what_changed():
            record.hash = self._recalculate_record_hash(record)

        return self._update(
            context, tables.records, record, exceptions.DuplicateRecord,
            exceptions.RecordNotFound)

    def delete_record(self, context, record_id):
        """
        Delete a record

        :param context: RPC Context.
        :param record_id: Record ID to delete
        """
        # Fetch the existing record, we'll need to return it.
        record = self._find_records(context, {'id': record_id}, one=True)
        return self._delete(context, tables.records, record,
                            exceptions.RecordNotFound)

    def count_records(self, context, criterion=None):
        """
        Count records

        :param context: RPC Context.
        :param criterion: Criteria to filter by.
        """
        # Ensure that we return only active records
        rjoin = tables.records.join(
            tables.zones,
            tables.records.c.zone_id == tables.zones.c.id)

        query = (
            select(func.count(tables.records.c.id)).
            select_from(rjoin).
            where(tables.zones.c.deleted == '0')
        )

        query = self._apply_criterion(tables.records, query, criterion)
        query = self._apply_tenant_criteria(context, tables.records, query)
        query = self._apply_deleted_criteria(context, tables.records, query)

        with sql.get_read_session() as session:
            resultproxy = session.execute(query)
            result = resultproxy.fetchone()

        if result is None:
            return 0

        return result[0]

    # Blacklist Methods
    def _find_blacklists(self, context, criterion, one=False, marker=None,
                         limit=None, sort_key=None, sort_dir=None):
        return self._find(
            context, tables.blacklists, objects.Blacklist,
            objects.BlacklistList, exceptions.BlacklistNotFound, criterion,
            one, marker, limit, sort_key, sort_dir)

    def create_blacklist(self, context, blacklist):
        """
        Create a Blacklist.

        :param context: RPC Context.
        :param blacklist: Blacklist object with the values to be created.
        """
        return self._create(
            tables.blacklists, blacklist, exceptions.DuplicateBlacklist)

    def get_blacklist(self, context, blacklist_id):
        """
        Get a Blacklist via ID.

        :param context: RPC Context.
        :param blacklist_id: Blacklist ID to get.
        """
        return self._find_blacklists(context, {'id': blacklist_id}, one=True)

    def find_blacklists(self, context, criterion=None, marker=None, limit=None,
                        sort_key=None, sort_dir=None):
        """
        Find Blacklists

        :param context: RPC Context.
        :param criterion: Criteria to filter by.
        :param marker: Resource ID from which after the requested page will
                       start after
        :param limit: Integer limit of objects of the page size after the
                      marker
        :param sort_key: Key from which to sort after.
        :param sort_dir: Direction to sort after using sort_key.
        """
        return self._find_blacklists(context, criterion, marker=marker,
                                     limit=limit, sort_key=sort_key,
                                     sort_dir=sort_dir)

    def find_blacklist(self, context, criterion):
        """
        Find a single Blacklist.

        :param context: RPC Context.
        :param criterion: Criteria to filter by.
        """
        return self._find_blacklists(context, criterion, one=True)

    def update_blacklist(self, context, blacklist):
        """
        Update a Blacklist

        :param context: RPC Context.
        :param blacklist: Blacklist to update.
        """
        return self._update(
            context, tables.blacklists, blacklist,
            exceptions.DuplicateBlacklist, exceptions.BlacklistNotFound)

    def delete_blacklist(self, context, blacklist_id):
        """
        Delete a Blacklist via ID.

        :param context: RPC Context.
        :param blacklist_id: Delete a Blacklist via ID
        """
        # Fetch the existing blacklist, we'll need to return it.
        blacklist = self._find_blacklists(
            context, {'id': blacklist_id}, one=True)

        return self._delete(context, tables.blacklists, blacklist,
                            exceptions.BlacklistNotFound)

    # Pool methods
    def _find_pools(self, context, criterion, one=False, marker=None,
                    limit=None, sort_key=None, sort_dir=None):
        pools = self._find(context, tables.pools, objects.Pool,
                           objects.PoolList, exceptions.PoolNotFound,
                           criterion, one, marker, limit, sort_key,
                           sort_dir)

        # Load Relations
        def _load_relations(pool):
            pool.attributes = self._find_pool_attributes(
                context, {'pool_id': pool.id})

            pool.ns_records = self._find_pool_ns_records(
                context, {'pool_id': pool.id})

            pool.nameservers = self._find_pool_nameservers(
                context, {'pool_id': pool.id})

            pool.targets = self._find_pool_targets(
                context, {'pool_id': pool.id})

            pool.also_notifies = self._find_pool_also_notifies(
                context, {'pool_id': pool.id})

            try:
                catalog_zone = self.get_catalog_zone(context, pool)

                try:
                    tsigkey = self.find_tsigkey(
                        context, criterion={'resource_id': catalog_zone.id})
                except exceptions.TsigKeyNotFound:
                    tsigkey = None

                secret = tsigkey.secret if tsigkey is not None else None
                algorithm = tsigkey.algorithm if tsigkey is not None else None

                pool.catalog_zone = objects.PoolCatalogZone(
                    catalog_zone_fqdn=catalog_zone.name,
                    catalog_zone_refresh=catalog_zone.refresh,
                    catalog_zone_tsig_key=secret,
                    catalog_zone_tsig_algorithm=algorithm,
                )
            except exceptions.ZoneNotFound:
                pool.catalog_zone = None

            pool.obj_reset_changes(['attributes', 'ns_records', 'nameservers',
                                    'targets', 'also_notifies', 'catalog_zone']
                                   )

        if one:
            _load_relations(pools)
        else:
            for pool in pools:
                _load_relations(pool)

        return pools

    def create_pool(self, context, pool):
        """
        Create a Pool.

        :param context: RPC Context.
        :param pool: Pool object with the values to be created.
        """
        pool = self._create(
            tables.pools, pool, exceptions.DuplicatePool,
            ['attributes', 'ns_records', 'nameservers', 'targets',
             'also_notifies', 'catalog_zone'])

        if pool.obj_attr_is_set('attributes'):
            for pool_attribute in pool.attributes:
                self.create_pool_attribute(context, pool.id, pool_attribute)
        else:
            pool.attributes = objects.PoolAttributeList()

        if pool.obj_attr_is_set('ns_records'):
            for ns_record in pool.ns_records:
                self.create_pool_ns_record(context, pool.id, ns_record)
        else:
            pool.ns_records = objects.PoolNsRecordList()

        if pool.obj_attr_is_set('nameservers'):
            for nameserver in pool.nameservers:
                self.create_pool_nameserver(context, pool.id, nameserver)
        else:
            pool.nameservers = objects.PoolNameserverList()

        if pool.obj_attr_is_set('targets'):
            for target in pool.targets:
                self.create_pool_target(context, pool.id, target)
        else:
            pool.targets = objects.PoolTargetList()

        if pool.obj_attr_is_set('also_notifies'):
            for also_notify in pool.also_notifies:
                self.create_pool_also_notify(context, pool.id, also_notify)
        else:
            pool.also_notifies = objects.PoolAlsoNotifyList()

        pool.obj_reset_changes(['attributes', 'ns_records', 'nameservers',
                                'targets', 'also_notifies'])

        self._ensure_catalog_zone_config(context, pool)

        return pool

    def get_pool(self, context, pool_id):
        """
        Get a Pool via the id

        :param context: RPC Context.
        :param pool_id: The ID of the pool to get
        """
        return self._find_pools(context, {'id': pool_id}, one=True)

    def find_pools(self, context, criterion=None, marker=None,
                   limit=None, sort_key=None, sort_dir=None):
        """
        Find all Pools

        :param context: RPC Context.
        :param criterion: Criteria by which to filter
        :param marker: Resource ID used by paging. The next page will start
                       at the next resource after the marker
        :param limit: Integer limit of objects on the page
        :param sort_key: Key used to sort the returned list
        :param sort_dir: Directions to sort after using sort_key
        """
        return self._find_pools(context, criterion, marker=marker,
                                limit=limit, sort_key=sort_key,
                                sort_dir=sort_dir)

    def find_pool(self, context, criterion):
        """
        Find a single Pool.

        :param context: RPC Context.
        :param criterion: Criteria to filter by.
        """
        return self._find_pools(context, criterion, one=True)

    def update_pool(self, context, pool):
        """
        Update the specified pool

        :param context: RPC Context.
        :param pool: Pool to update.
        """
        pool = self._update(context, tables.pools, pool,
                            exceptions.DuplicatePool, exceptions.PoolNotFound,
                            ['attributes', 'ns_records', 'nameservers',
                             'targets', 'also_notifies', 'catalog_zone'])

        for attribute_name in ('attributes', 'ns_records', 'nameservers',
                               'targets', 'also_notifies'):
            if pool.obj_attr_is_set(attribute_name):
                self._update_pool_items(context, pool, attribute_name)

        # Call get_pool to get the ids of all the attributes/ns_records
        # refreshed in the pool object
        updated_pool = self.get_pool(context, pool.id)

        self._ensure_catalog_zone_config(context, pool)

        return updated_pool

    def delete_pool(self, context, pool_id):
        """
        Delete the pool with the matching id

        :param context: RPC Context.
        :param pool_id: The ID of the pool to be deleted
        """
        pool = self._find_pools(context, {'id': pool_id}, one=True)

        try:
            catalog_zone = self.get_catalog_zone(context, pool)

            try:
                catalog_zone_tsig = self.find_tsigkey(
                    context, criterion={'resource_id': catalog_zone.id})
                self.delete_tsigkey(context, catalog_zone_tsig.id)
            except exceptions.TsigKeyNotFound:
                pass

            self.delete_zone(context, catalog_zone.id)
        except exceptions.ZoneNotFound:
            pass

        return self._delete(context, tables.pools, pool,
                            exceptions.PoolNotFound)

    # Pool attribute methods
    def _find_pool_attributes(self, context, criterion, one=False, marker=None,
                              limit=None, sort_key=None, sort_dir=None):
        return self._find(context, tables.pool_attributes,
                          objects.PoolAttribute, objects.PoolAttributeList,
                          exceptions.PoolAttributeNotFound, criterion, one,
                          marker, limit, sort_key, sort_dir)

    def create_pool_attribute(self, context, pool_id, pool_attribute):
        """
        Create a PoolAttribute.

        :param context: RPC Context.
        :param pool_id: The ID of the pool to which the attribute belongs.
        :param pool_attribute: PoolAttribute object with the values created.
        """
        pool_attribute.pool_id = pool_id

        return self._create(tables.pool_attributes, pool_attribute,
                            exceptions.DuplicatePoolAttribute)

    def get_pool_attribute(self, context, pool_attribute_id):
        """
        Get a PoolAttribute via the ID

        :param context: RPC Context.
        :param pool_attribute_id: The ID of the PoolAttribute to get
        """
        return self._find_pool_attributes(
            context, {'id': pool_attribute_id}, one=True)

    def find_pool_attributes(self, context, criterion=None, marker=None,
                             limit=None, sort_key=None, sort_dir=None):
        """
        Find all PoolAttributes

        :param context: RPC Context
        :param criterion: Criteria by which to filer
        :param marker: Resource ID used by paging. The next page will start
                       at the next resource after the marker
        :param limit: Integer limit of objects on the page
        :param sort_key: Key used to sort the returned list
        :param sort_dir: Directions to sort after using sort_key
        """
        return self._find_pool_attributes(context, criterion, marker=marker,
                                          limit=limit, sort_key=sort_key,
                                          sort_dir=sort_dir)

    def find_pool_attribute(self, context, criterion):
        """
        Find a single PoolAttribute

        :param context: RPC Context.
        :param criterion: Criteria to filter by.
        """
        return self._find_pool_attributes(context, criterion, one=True)

    def update_pool_attribute(self, context, pool_attribute):
        """
        Update the specified pool

        :param context: RPC Context.
        :param pool_attribute: PoolAttribute to update
        """
        return self._update(context, tables.pool_attributes, pool_attribute,
                            exceptions.DuplicatePoolAttribute,
                            exceptions.PoolAttributeNotFound)

    def delete_pool_attribute(self, context, pool_attribute_id):
        """
        Delete the pool with the matching id

        :param context: RPC Context.
        :param pool_attribute_id: The ID of the PoolAttribute to be deleted
        """
        pool_attribute = self._find_pool_attributes(
            context, {'id': pool_attribute_id}, one=True)
        deleted_pool_attribute = self._delete(
            context, tables.pool_attributes, pool_attribute,
            exceptions.PoolAttributeNotFound)

        return deleted_pool_attribute

    # Pool ns_record methods
    def _find_pool_ns_records(self, context, criterion, one=False, marker=None,
                              limit=None, sort_key=None, sort_dir=None):
        return self._find(context, tables.pool_ns_records,
                          objects.PoolNsRecord, objects.PoolNsRecordList,
                          exceptions.PoolNsRecordNotFound, criterion, one,
                          marker, limit, sort_key, sort_dir)

    def create_pool_ns_record(self, context, pool_id, pool_ns_record):
        pool_ns_record.pool_id = pool_id

        return self._create(tables.pool_ns_records, pool_ns_record,
                            exceptions.DuplicatePoolNsRecord)

    def update_pool_ns_record(self, context, pool_ns_record):
        return self._update(context, tables.pool_ns_records, pool_ns_record,
                            exceptions.DuplicatePoolNsRecord,
                            exceptions.PoolNsRecordNotFound)

    def delete_pool_ns_record(self, context, pool_ns_record_id):
        pool_ns_record = self._find_pool_ns_records(
            context, {'id': pool_ns_record_id}, one=True)
        deleted_pool_ns_record = self._delete(
            context, tables.pool_ns_records, pool_ns_record,
            exceptions.PoolNsRecordNotFound)

        return deleted_pool_ns_record

    # PoolNameserver methods
    def _find_pool_nameservers(self, context, criterion, one=False,
                               marker=None, limit=None, sort_key=None,
                               sort_dir=None):
        return self._find(context, tables.pool_nameservers,
                          objects.PoolNameserver, objects.PoolNameserverList,
                          exceptions.PoolNameserverNotFound, criterion, one,
                          marker, limit, sort_key, sort_dir)

    def create_pool_nameserver(self, context, pool_id, pool_nameserver):
        pool_nameserver.pool_id = pool_id

        return self._create(tables.pool_nameservers, pool_nameserver,
                            exceptions.DuplicatePoolNameserver)

    def get_pool_nameserver(self, context, pool_nameserver_id):
        return self._find_pool_nameservers(
            context, {'id': pool_nameserver_id}, one=True)

    def find_pool_nameservers(self, context, criterion=None, marker=None,
                              limit=None, sort_key=None, sort_dir=None):
        return self._find_pool_nameservers(context, criterion, marker=marker,
                                           limit=limit, sort_key=sort_key,
                                           sort_dir=sort_dir)

    def find_pool_nameserver(self, context, criterion):
        return self._find_pool_nameservers(context, criterion, one=True)

    def update_pool_nameserver(self, context, pool_nameserver):
        return self._update(context, tables.pool_nameservers, pool_nameserver,
                            exceptions.DuplicatePoolNameserver,
                            exceptions.PoolNameserverNotFound)

    def delete_pool_nameserver(self, context, pool_nameserver_id):
        pool_nameserver = self._find_pool_nameservers(
            context, {'id': pool_nameserver_id}, one=True)
        deleted_pool_nameserver = self._delete(
            context, tables.pool_nameservers, pool_nameserver,
            exceptions.PoolNameserverNotFound)

        return deleted_pool_nameserver

    # PoolTarget methods
    def _find_pool_targets(self, context, criterion, one=False, marker=None,
                           limit=None, sort_key=None, sort_dir=None):
        pool_targets = self._find(
            context, tables.pool_targets, objects.PoolTarget,
            objects.PoolTargetList, exceptions.PoolTargetNotFound,
            criterion, one, marker, limit, sort_key,
            sort_dir)

        # Load Relations
        def _load_relations(pool_target):
            pool_target.options = self._find_pool_target_options(
                context, {'pool_target_id': pool_target.id})

            pool_target.masters = self._find_pool_target_masters(
                context, {'pool_target_id': pool_target.id})

            pool_target.obj_reset_changes(['options', 'masters'])

        if one:
            _load_relations(pool_targets)
        else:
            for pool_target in pool_targets:
                _load_relations(pool_target)

        return pool_targets

    def create_pool_target(self, context, pool_id, pool_target):
        pool_target.pool_id = pool_id

        pool_target = self._create(
            tables.pool_targets, pool_target, exceptions.DuplicatePoolTarget,
            ['options', 'masters'])

        if pool_target.obj_attr_is_set('options'):
            for pool_target_option in pool_target.options:
                self.create_pool_target_option(
                    context, pool_target.id, pool_target_option)
        else:
            pool_target.options = objects.PoolTargetOptionList()

        if pool_target.obj_attr_is_set('masters'):
            for pool_target_master in pool_target.masters:
                self.create_pool_target_master(
                    context, pool_target.id, pool_target_master)
        else:
            pool_target.masters = objects.PoolTargetMasterList()

        pool_target.obj_reset_changes(['options', 'masters'])

        return pool_target

    def get_pool_target(self, context, pool_target_id):
        return self._find_pool_targets(
            context, {'id': pool_target_id}, one=True)

    def find_pool_targets(self, context, criterion=None, marker=None,
                          limit=None, sort_key=None, sort_dir=None):
        return self._find_pool_targets(context, criterion, marker=marker,
                                       limit=limit, sort_key=sort_key,
                                       sort_dir=sort_dir)

    def find_pool_target(self, context, criterion):
        return self._find_pool_targets(context, criterion, one=True)

    def _update_pool_items(self, context, pool, attribute_name):
        """Update attributes beloging to a pool
        """
        assert attribute_name in ('attributes', 'ns_records', 'nameservers',
                                  'targets', 'also_notifies')

        # Gather the pool ID's we have
        finder = getattr(self, '_find_pool_%s' % attribute_name)
        have_items = set()
        for r in finder(context, {'pool_id': pool.id}):
            have_items.add(r.id)

        # Prep some lists of changes
        keep_items = set()
        create_items = []
        update_items = []

        items = []
        if pool.obj_attr_is_set(attribute_name):
            for r in getattr(pool, attribute_name).objects:
                items.append(r)

        # Determine what to change
        for item in items:
            keep_items.add(item.id)
            try:
                item.obj_get_original_value('id')
            except KeyError:
                create_items.append(item)
            else:
                update_items.append(item)

        # NOTE: Since we're dealing with mutable objects, the return value
        #       of create/update/delete option is not needed. The
        #       original item will be mutated in place on the input
        #       "pool.options" list.

        # singular: attributes -> attribute, 'notify' is as a corner case
        if attribute_name == 'also_notifies':
            singular = 'also_notify'
        else:
            singular = attribute_name[:-1]

        # Delete items
        fn = getattr(self, 'delete_pool_%s' % singular)
        for item_id in have_items - keep_items:
            fn(context, item_id)

        # Update items
        fn = getattr(self, 'update_pool_%s' % singular)
        for item in update_items:
            fn(context, item)

        # Create items
        fn = getattr(self, 'create_pool_%s' % singular)
        for item in create_items:
            fn(context, pool.id, item)

    def _update_pool_target_items(self, context, pool_target, attribute_name):
        """Update attributes beloging to a pool target
        """
        assert attribute_name in ('options', 'masters')

        # Gather the pool ID's we have
        finder = getattr(self, '_find_pool_target_%s' % attribute_name)
        have_items = set()
        for r in finder(context, {'pool_target_id': pool_target.id}):
            have_items.add(r.id)

        # Prep some lists of changes
        keep_items = set()
        create_items = []
        update_items = []

        items = []
        if pool_target.obj_attr_is_set(attribute_name):
            for r in getattr(pool_target, attribute_name).objects:
                items.append(r)

        # Determine what to change
        for item in items:
            keep_items.add(item.id)
            try:
                item.obj_get_original_value('id')
            except KeyError:
                create_items.append(item)
            else:
                update_items.append(item)

        # NOTE: Since we're dealing with mutable objects, the return value
        #       of create/update/delete option is not needed. The
        #       original item will be mutated in place on the input
        #       "pool.options" list.

        # singular: options -> option
        singular = attribute_name[:-1]

        # Delete items
        fn = getattr(self, 'delete_pool_target_%s' % singular)
        for item_id in have_items - keep_items:
            fn(context, item_id)

        # Update items
        fn = getattr(self, 'update_pool_target_%s' % singular)
        for item in update_items:
            fn(context, item)

        # Create items
        fn = getattr(self, 'create_pool_target_%s' % singular)
        for item in create_items:
            fn(context, pool_target.id, item)

    def update_pool_target(self, context, pool_target):
        pool_target = self._update(
            context, tables.pool_targets, pool_target,
            exceptions.DuplicatePoolTarget, exceptions.PoolTargetNotFound,
            ['options', 'masters'])

        for attribute_name in ('options', 'masters'):
            if pool_target.obj_attr_is_set(attribute_name):
                self._update_pool_target_items(context, pool_target,
                                               attribute_name)

        # Call get_pool to get the ids of all the attributes/ns_records
        # refreshed in the pool object
        updated_pool_target = self.get_pool_target(context, pool_target.id)

        return updated_pool_target

    def delete_pool_target(self, context, pool_target_id):
        pool_target = self._find_pool_targets(
            context, {'id': pool_target_id}, one=True)

        return self._delete(context, tables.pool_targets, pool_target,
                            exceptions.PoolTargetNotFound)

    # PoolTargetOption methods
    def _find_pool_target_options(self, context, criterion, one=False,
                                  marker=None, limit=None, sort_key=None,
                                  sort_dir=None):
        return self._find(
            context, tables.pool_target_options,
            objects.PoolTargetOption, objects.PoolTargetOptionList,
            exceptions.PoolTargetOptionNotFound, criterion, one,
            marker, limit, sort_key, sort_dir)

    def create_pool_target_option(self, context, pool_target_id,
                                  pool_target_option):
        pool_target_option.pool_target_id = pool_target_id

        return self._create(tables.pool_target_options, pool_target_option,
                            exceptions.DuplicatePoolTargetOption)

    def update_pool_target_option(self, context, pool_target_option):
        return self._update(
            context, tables.pool_target_options, pool_target_option,
            exceptions.DuplicatePoolTargetOption,
            exceptions.PoolTargetOptionNotFound)

    def delete_pool_target_option(self, context, pool_target_option_id):
        pool_target_option = self._find_pool_target_options(
            context, {'id': pool_target_option_id}, one=True)
        deleted_pool_target_option = self._delete(
            context, tables.pool_target_options, pool_target_option,
            exceptions.PoolTargetOptionNotFound)

        return deleted_pool_target_option

    # PoolTargetMaster methods
    def _find_pool_target_masters(self, context, criterion, one=False,
                                  marker=None, limit=None, sort_key=None,
                                  sort_dir=None):
        return self._find(
            context, tables.pool_target_masters,
            objects.PoolTargetMaster, objects.PoolTargetMasterList,
            exceptions.PoolTargetMasterNotFound, criterion, one,
            marker, limit, sort_key, sort_dir)

    def create_pool_target_master(self, context, pool_target_id,
                                  pool_target_master):
        pool_target_master.pool_target_id = pool_target_id

        return self._create(tables.pool_target_masters, pool_target_master,
                            exceptions.DuplicatePoolTargetMaster)

    def update_pool_target_master(self, context, pool_target_master):
        return self._update(
            context, tables.pool_target_masters, pool_target_master,
            exceptions.DuplicatePoolTargetMaster,
            exceptions.PoolTargetMasterNotFound)

    def delete_pool_target_master(self, context, pool_target_master_id):
        pool_target_master = self._find_pool_target_masters(
            context, {'id': pool_target_master_id}, one=True)
        deleted_pool_target_master = self._delete(
            context, tables.pool_target_masters, pool_target_master,
            exceptions.PoolTargetMasterNotFound)

        return deleted_pool_target_master

    # PoolAlsoNotify methods
    def _find_pool_also_notifies(self, context, criterion, one=False,
                                 marker=None, limit=None, sort_key=None,
                                 sort_dir=None):
        return self._find(context, tables.pool_also_notifies,
                          objects.PoolAlsoNotify, objects.PoolAlsoNotifyList,
                          exceptions.PoolAlsoNotifyNotFound, criterion, one,
                          marker, limit, sort_key, sort_dir)

    def create_pool_also_notify(self, context, pool_id, pool_also_notify):
        pool_also_notify.pool_id = pool_id

        return self._create(tables.pool_also_notifies, pool_also_notify,
                            exceptions.DuplicatePoolAlsoNotify)

    def get_pool_also_notify(self, context, pool_also_notify_id):
        return self._find_pool_also_notifies(
            context, {'id': pool_also_notify_id}, one=True)

    def find_pool_also_notifies(self, context, criterion=None, marker=None,
                                limit=None, sort_key=None, sort_dir=None):
        return self._find_pool_also_notifies(context, criterion, marker=marker,
                                             limit=limit, sort_key=sort_key,
                                             sort_dir=sort_dir)

    def find_pool_also_notify(self, context, criterion):
        return self._find_pool_also_notifies(context, criterion, one=True)

    def update_pool_also_notify(self, context, pool_also_notify):
        return self._update(
            context, tables.pool_also_notifies, pool_also_notify,
            exceptions.DuplicatePoolAlsoNotify,
            exceptions.PoolAlsoNotifyNotFound)

    def delete_pool_also_notify(self, context, pool_also_notify_id):
        pool_also_notify = self._find_pool_also_notifies(
            context, {'id': pool_also_notify_id}, one=True)
        deleted_pool_also_notify = self._delete(
            context, tables.pool_also_notifies, pool_also_notify,
            exceptions.PoolAlsoNotifyNotFound)

        return deleted_pool_also_notify

    # Zone Transfer Methods
    def _find_zone_transfer_requests(self, context, criterion, one=False,
                                     marker=None, limit=None, sort_key=None,
                                     sort_dir=None):

        table = tables.zone_transfer_requests

        ljoin = tables.zone_transfer_requests.join(
            tables.zones,
            tables.zone_transfer_requests.c.zone_id == tables.zones.c.id)

        query = select(
            table, tables.zones.c.name.label('zone_name')
        ).select_from(ljoin)

        if not context.all_tenants:
            # If we have a system scoped token with no project_id and
            # all_tenants was not used, we don't know what records to return,
            # so return an empty list.
            if not context.project_id:
                if one:
                    return objects.ZoneTransferRequest()
                return objects.ZoneTransferRequestList()

            query = query.where(or_(
                table.c.tenant_id == context.project_id,
                table.c.target_tenant_id == context.project_id))

        return self._find(
            context, table, objects.ZoneTransferRequest,
            objects.ZoneTransferRequestList,
            exceptions.ZoneTransferRequestNotFound,
            criterion,
            one=one, marker=marker, limit=limit, sort_dir=sort_dir,
            sort_key=sort_key, query=query,
            apply_tenant_criteria=False
        )

    def create_zone_transfer_request(self, context, zone_transfer_request):

        try:
            criterion = {'zone_id': zone_transfer_request.zone_id,
                         'status': 'ACTIVE'}
            self.find_zone_transfer_request(
                context, criterion)
        except exceptions.ZoneTransferRequestNotFound:
            return self._create(
                tables.zone_transfer_requests,
                zone_transfer_request,
                exceptions.DuplicateZoneTransferRequest)
        else:
            raise exceptions.DuplicateZoneTransferRequest()

    def find_zone_transfer_requests(self, context, criterion=None,
                                    marker=None, limit=None, sort_key=None,
                                    sort_dir=None):

        return self._find_zone_transfer_requests(
            context, criterion, marker=marker,
            limit=limit, sort_key=sort_key,
            sort_dir=sort_dir)

    def get_zone_transfer_request(self, context, zone_transfer_request_id):
        request = self._find_zone_transfer_requests(
            context,
            {'id': zone_transfer_request_id},
            one=True)

        return request

    def find_zone_transfer_request(self, context, criterion):

        return self._find_zone_transfer_requests(context, criterion, one=True)

    def update_zone_transfer_request(self, context, zone_transfer_request):

        zone_transfer_request.obj_reset_changes('zone_name')

        updated_zt_request = self._update(
            context,
            tables.zone_transfer_requests,
            zone_transfer_request,
            exceptions.DuplicateZoneTransferRequest,
            exceptions.ZoneTransferRequestNotFound,
            skip_values=['zone_name'])

        return updated_zt_request

    def delete_zone_transfer_request(self, context, zone_transfer_request_id):

        zone_transfer_request = self._find_zone_transfer_requests(
            context,
            {'id': zone_transfer_request_id},
            one=True)

        return self._delete(
            context,
            tables.zone_transfer_requests,
            zone_transfer_request,
            exceptions.ZoneTransferRequestNotFound)

    def count_zone_transfer_accept(self, context, criterion=None):
        query = select(func.count(tables.zone_transfer_accepts.c.id))
        query = self._apply_criterion(
            tables.zone_transfer_accepts, query, criterion
        )
        query = self._apply_deleted_criteria(
            context, tables.zone_transfer_accepts, query
        )

        with sql.get_read_session() as session:
            resultproxy = session.execute(query)
            result = resultproxy.fetchone()

        if result is None:
            return 0

        return result[0]

    def _find_zone_transfer_accept(self, context, criterion, one=False,
                                   marker=None, limit=None, sort_key=None,
                                   sort_dir=None):

        zone_transfer_accept = self._find(
            context, tables.zone_transfer_accepts,
            objects.ZoneTransferAccept,
            objects.ZoneTransferAcceptList,
            exceptions.ZoneTransferAcceptNotFound, criterion,
            one, marker, limit, sort_key, sort_dir)

        if not one:
            zone_transfer_accept.total_count = self.count_zone_transfer_accept(
                context, criterion)

        return zone_transfer_accept

    def create_zone_transfer_accept(self, context, zone_transfer_accept):

        return self._create(
            tables.zone_transfer_accepts,
            zone_transfer_accept,
            exceptions.DuplicateZoneTransferAccept)

    def find_zone_transfer_accepts(self, context, criterion=None,
                                   marker=None, limit=None, sort_key=None,
                                   sort_dir=None):
        return self._find_zone_transfer_accept(
            context, criterion, marker=marker, limit=limit, sort_key=sort_key,
            sort_dir=sort_dir)

    def get_zone_transfer_accept(self, context, zone_transfer_accept_id):
        return self._find_zone_transfer_accept(
            context,
            {'id': zone_transfer_accept_id},
            one=True)

    def find_zone_transfer_accept(self, context, criterion):
        return self._find_zone_transfer_accept(
            context,
            criterion,
            one=True)

    def update_zone_transfer_accept(self, context, zone_transfer_accept):

        return self._update(
            context,
            tables.zone_transfer_accepts,
            zone_transfer_accept,
            exceptions.DuplicateZoneTransferAccept,
            exceptions.ZoneTransferAcceptNotFound)

    def delete_zone_transfer_accept(self, context, zone_transfer_accept_id):

        zone_transfer_accept = self._find_zone_transfer_accept(
            context,
            {'id': zone_transfer_accept_id},
            one=True)

        return self._delete(
            context,
            tables.zone_transfer_accepts,
            zone_transfer_accept,
            exceptions.ZoneTransferAcceptNotFound)

    # Zone Import Methods
    def _find_zone_imports(self, context, criterion, one=False, marker=None,
                           limit=None, sort_key=None, sort_dir=None):
        if not criterion:
            criterion = {}
        criterion['task_type'] = 'IMPORT'
        zone_imports = self._find(
            context, tables.zone_tasks, objects.ZoneImport,
            objects.ZoneImportList, exceptions.ZoneImportNotFound, criterion,
            one, marker, limit, sort_key, sort_dir)

        if not one:
            zone_imports.total_count = self.count_zone_tasks(
                context, criterion)

        return zone_imports

    def create_zone_import(self, context, zone_import):
        """
        Create a Zone Import.

        :param context: RPC Context.
        :param zone_import: Zone Import object with the values to be created.
        """
        return self._create(
            tables.zone_tasks, zone_import, exceptions.DuplicateZoneImport)

    def get_zone_import(self, context, zone_import_id):
        """
        Get a Zone Import via ID.

        :param context: RPC Context.
        :param zone_import_id: Zone Import ID to get.
        """
        return self._find_zone_imports(context, {'id': zone_import_id},
                                       one=True)

    def find_zone_imports(self, context, criterion=None, marker=None,
                          limit=None, sort_key=None, sort_dir=None):
        """
        Find Zone Imports

        :param context: RPC Context.
        :param criterion: Criteria to filter by.
        :param marker: Resource ID from which after the requested page will
                       start after
        :param limit: Integer limit of objects of the page size after the
                      marker
        :param sort_key: Key from which to sort after.
        :param sort_dir: Direction to sort after using sort_key.
        """
        return self._find_zone_imports(context, criterion, marker=marker,
                                       limit=limit, sort_key=sort_key,
                                       sort_dir=sort_dir)

    def find_zone_import(self, context, criterion):
        """
        Find a single Zone Import.

        :param context: RPC Context.
        :param criterion: Criteria to filter by.
        """
        return self._find_zone_imports(context, criterion, one=True)

    def update_zone_import(self, context, zone_import):
        """
        Update a Zone Import

        :param context: RPC Context.
        :param zone_import: Zone Import to update.
        """
        return self._update(
            context, tables.zone_tasks, zone_import,
            exceptions.DuplicateZoneImport, exceptions.ZoneImportNotFound)

    def delete_zone_import(self, context, zone_import_id):
        """
        Delete a Zone Import via ID.

        :param context: RPC Context.
        :param zone_import_id: Delete a Zone Import via ID
        """
        # Fetch the existing zone_import, we'll need to return it.
        zone_import = self._find_zone_imports(context, {'id': zone_import_id},
                                              one=True)
        return self._delete(context, tables.zone_tasks, zone_import,
                            exceptions.ZoneImportNotFound)

    # Zone Export Methods
    def _find_zone_exports(self, context, criterion, one=False, marker=None,
                           limit=None, sort_key=None, sort_dir=None):
        if not criterion:
            criterion = {}
        criterion['task_type'] = 'EXPORT'
        zone_exports = self._find(
            context, tables.zone_tasks, objects.ZoneExport,
            objects.ZoneExportList, exceptions.ZoneExportNotFound, criterion,
            one, marker, limit, sort_key, sort_dir)
        if not one:
            zone_exports.total_count = self.count_zone_tasks(
                context, criterion)

        return zone_exports

    def create_zone_export(self, context, zone_export):
        """
        Create a Zone Export.

        :param context: RPC Context.
        :param zone_export: Zone Export object with the values to be created.
        """
        return self._create(
            tables.zone_tasks, zone_export, exceptions.DuplicateZoneExport)

    def get_zone_export(self, context, zone_export_id):
        """
        Get a Zone Export via ID.

        :param context: RPC Context.
        :param zone_export_id: Zone Export ID to get.
        """
        return self._find_zone_exports(context, {'id': zone_export_id},
                                       one=True)

    def find_zone_exports(self, context, criterion=None, marker=None,
                          limit=None, sort_key=None, sort_dir=None):
        """
        Find Zone Exports

        :param context: RPC Context.
        :param criterion: Criteria to filter by.
        :param marker: Resource ID from which after the requested page will
                       start after
        :param limit: Integer limit of objects of the page size after the
                      marker
        :param sort_key: Key from which to sort after.
        :param sort_dir: Direction to sort after using sort_key.
        """
        return self._find_zone_exports(context, criterion, marker=marker,
                                       limit=limit, sort_key=sort_key,
                                       sort_dir=sort_dir)

    def find_zone_export(self, context, criterion):
        """
        Find a single Zone Export.

        :param context: RPC Context.
        :param criterion: Criteria to filter by.
        """
        return self._find_zone_exports(context, criterion, one=True)

    def update_zone_export(self, context, zone_export):
        """
        Update a Zone Export

        :param context: RPC Context.
        :param zone_export: Zone Export to update.
        """
        return self._update(
            context, tables.zone_tasks, zone_export,
            exceptions.DuplicateZoneExport, exceptions.ZoneExportNotFound)

    def delete_zone_export(self, context, zone_export_id):
        """
        Delete a Zone Export via ID.

        :param context: RPC Context.
        :param zone_export_id: Delete a Zone Export via ID
        """
        # Fetch the existing zone_export, we'll need to return it.
        zone_export = self._find_zone_exports(context, {'id': zone_export_id},
                                              one=True)
        return self._delete(context, tables.zone_tasks, zone_export,
                            exceptions.ZoneExportNotFound)

    def count_zone_tasks(self, context, criterion=None):
        query = select(func.count(tables.zone_tasks.c.id))
        query = self._apply_criterion(tables.zone_tasks, query, criterion)
        query = self._apply_tenant_criteria(context, tables.zone_tasks, query)
        query = self._apply_deleted_criteria(context, tables.zone_tasks, query)

        with sql.get_read_session() as session:
            resultproxy = session.execute(query)
            result = resultproxy.fetchone()

        if result is None:
            return 0

        return result[0]

    # Service Status Methods
    def _find_service_statuses(self, context, criterion, one=False,
                               marker=None, limit=None, sort_key=None,
                               sort_dir=None):
        return self._find(
            context, tables.service_status, objects.ServiceStatus,
            objects.ServiceStatusList, exceptions.ServiceStatusNotFound,
            criterion, one, marker, limit, sort_key, sort_dir)

    def find_service_status(self, context, criterion):
        """
        Find a single Service Status.

        :param context: RPC Context.
        :param criterion: Criteria to filter by.
        """
        return self._find_service_statuses(context, criterion, one=True)

    def find_service_statuses(self, context, criterion=None, marker=None,
                              limit=None, sort_key=None, sort_dir=None):
        """
        Retrieve status for services

        :param context: RPC Context.
        :param criterion: Criteria to filter by.
        :param marker: Resource ID from which after the requested page will
                       start after
        :param limit: Integer limit of objects of the page size after the
                      marker
        :param sort_key: Key from which to sort after.
        :param sort_dir: Direction to sort after using sort_key.
        """
        return self._find_service_statuses(context, criterion, marker=marker,
                                           limit=limit, sort_key=sort_key,
                                           sort_dir=sort_dir)

    def create_service_status(self, context, service_status):
        """
        Create a Service status for a service.

        :param context: RPC Context.
        :param service_status: The status of a service.
        """
        return self._create(
            tables.service_status, service_status,
            exceptions.DuplicateServiceStatus)

    def update_service_status(self, context, service_status):
        """
        Update the Service status for a service.

        :param context: RPC Context.
        :param service_status: Set the status for a service.
        """
        return self._update(
            context, tables.service_status, service_status,
            exceptions.DuplicateServiceStatus,
            exceptions.ServiceStatusNotFound)

    # Reverse Name utils
    def _rname_check(self, criterion):
        # If the criterion has 'name' in it, switch it out for reverse_name
        if criterion is not None and criterion.get('name', '').startswith('*'):
            criterion['reverse_name'] = criterion.pop('name')[::-1]
        return criterion

    def _create_catalog_zone(self, pool):
        catalog_zone = objects.Zone(
            name=pool.catalog_zone.catalog_zone_fqdn,
            email=CONF[
                'service:central'].managed_resource_email,
            refresh=pool.catalog_zone.catalog_zone_refresh,
            serial=1,
            pool_id=pool.id,
            type=constants.ZONE_CATALOG
        )
        return catalog_zone

    def get_catalog_zone(self, context, pool):
        catalog_zone = self.find_zone(
            context, criterion={
                'pool_id': pool.id, 'type': constants.ZONE_CATALOG})
        return catalog_zone

    def _ensure_catalog_zone_config(self, context, pool):
        if (
            not pool.obj_attr_is_set("catalog_zone") or
            pool.catalog_zone is None
        ):
            return

        try:
            self.get_catalog_zone(context, pool)
        except exceptions.ZoneNotFound:
            catalog_zone = self._create_catalog_zone(pool)
            self.create_zone(context, catalog_zone)

        self._ensure_catalog_zone_consistent(context, pool)

    def _ensure_catalog_zone_consistent(self, context, pool):
        """
        Ensure a catalog zone's data as defined in pools.yaml is consistent
        with its values in the database.

        :param context: RPC Context.
        :param pool: The pool to ensure catalog zone consistency for.
        """
        if not pool.obj_attr_is_set("catalog_zone") or not pool.catalog_zone:
            return

        catalog_zone = self.get_catalog_zone(context, pool)
        catalog_zone.attributes = self._find_zone_attributes(
                context, {'zone_id': catalog_zone.id})

        catalog_zone = self._ensure_catalog_zone_info_consistent(
            context, catalog_zone, pool.catalog_zone) or catalog_zone
        self._ensure_catalog_zone_soa_consistent(context, catalog_zone, pool)
        self._ensure_catalog_zone_tsig_data_consistent(
            context, catalog_zone, pool.catalog_zone)

    def _ensure_catalog_zone_info_consistent(
            self, context, catalog_zone, values):
        """
        Ensure a catalog zone's FQDN and refresh interval as defined in
        pools.yaml are consistent with their values in the database.

        :param context: RPC Context.
        :param catalog_zone: The catalog zone to ensure consistency for.
        :param values: The catalog zone values as defined in pools.yaml.
        """
        if (
            not catalog_zone.attributes or
            catalog_zone.attributes.get('catalog_zone_fqdn') !=
            values.catalog_zone_fqdn or
            catalog_zone.attributes.get('catalog_zone_refresh') !=
            values.catalog_zone_refresh
        ):
            catalog_zone.attributes = objects.ZoneAttributeList()
            catalog_zone_fqdn = objects.ZoneAttribute()
            catalog_zone_fqdn.zone_id = catalog_zone.id
            catalog_zone_fqdn.key = 'catalog_zone_fqdn'
            catalog_zone_fqdn.value = values.catalog_zone_fqdn

            catalog_zone_refresh = objects.ZoneAttribute()
            catalog_zone_refresh.zone_id = catalog_zone.id
            catalog_zone_refresh.key = 'catalog_zone_refresh'
            catalog_zone_refresh.value = values.catalog_zone_refresh
            catalog_zone.attributes.append(catalog_zone_fqdn)
            catalog_zone.attributes.append(catalog_zone_refresh)

            return self.update_zone(context, catalog_zone)

    def _ensure_catalog_zone_soa_consistent(self, context, catalog_zone, pool):
        """
        Ensure a catalog zone's SOA based on the values in pools.yaml is
        consistent with its values in the database.

        :param context: RPC Context.
        :param catalog_zone: The catalog zone to ensure consistency for.
        :param pool: The catalog_zone's pool.
        """
        soa_record = objects.RecordList()
        soa_record.append(
            objects.Record(
                    data=f'{pool.ns_records[0]["hostname"]} '
                    f'{catalog_zone.attributes.get("catalog_zone_fqdn")} '
                    f'{catalog_zone.serial} '
                    f'{catalog_zone.attributes.get("catalog_zone_refresh")} '
                    f'{catalog_zone.retry} '
                    '2147483646 '
                    f'{catalog_zone.minimum}'
            )
        )
        soa = objects.RecordSet(
            name=catalog_zone.name,
            type='SOA',
            records=soa_record
        )

        try:
            soa_db = self.find_recordset(
                context, criterion={'zone_id': catalog_zone.id, 'type': 'SOA'})
            soa_db.name = catalog_zone.name
            soa_db.records = soa_record
            self.update_recordset(context, soa_db)
        except exceptions.RecordSetNotFound:
            self.create_recordset(context, catalog_zone.id, soa)

    def _ensure_catalog_zone_tsig_data_consistent(
            self, context, catalog_zone, values):
        """
        Ensure a catalog zone's TSIG key and TSIG algorithm as defined in
        pools.yaml are consistent with their values in the database.

        :param context: RPC Context.
        :param catalog_zone: The catalog zone to ensure consistency for.
        :param values: The catalog zone values as defined in pools.yaml.
        """
        if (
            not values.catalog_zone_tsig_key or not
            values.catalog_zone_tsig_algorithm
        ):
            return

        tsig_key = values.catalog_zone_tsig_key
        tsig_algorithm = values.catalog_zone_tsig_algorithm

        try:
            tsigkey = self.find_tsigkey(
                context, criterion={'resource_id': catalog_zone.id})

            if (
                tsigkey.name !=
                    catalog_zone.attributes.get('catalog_zone_fqdn') or
                    tsigkey.secret != tsig_key or
                    tsigkey.algorithm != tsig_algorithm
            ):
                tsigkey.name = catalog_zone.attributes.get('catalog_zone_fqdn')
                tsigkey.secret = tsig_key
                tsigkey.algorithm = tsig_algorithm
                self.update_tsigkey(context, tsigkey)
        except exceptions.TsigKeyNotFound:
            tsigkey = objects.TsigKey(
                name=catalog_zone.attributes.get('catalog_zone_fqdn'),
                secret=tsig_key,
                algorithm=tsig_algorithm,
                scope='ZONE',
                resource_id=catalog_zone.id,
            )
            tsigkey = DesignateAdapter.parse(
                'API_v2', tsigkey, objects.TsigKey())
            tsigkey.validate()
            self.create_tsigkey(context, tsigkey)

    def get_catalog_zone_records(self, context, pool):
        catalog_zone = self.get_catalog_zone(context, pool)
        zones = self.find_zones(
            context, criterion={'pool_id': pool.id, 'type': '!CATALOG'})
        soa_record = self.find_recordset(
            context, criterion={'zone_id': catalog_zone.id, 'type': 'SOA'})
        records = []

        # Catalog zones require one NS record using NSDNAME 'invalid.'
        # per RFC 9432
        ns_record = objects.RecordList()
        ns_record.append(objects.Record(data='invalid.'))
        records.append(
            objects.RecordSet(
                name=pool.catalog_zone.catalog_zone_fqdn,
                type='NS',
                records=ns_record
            )
        )

        # Catalog zones require a TXT record with the schema version,
        # currently '2' per RFC 9432
        txt_record = objects.RecordList()
        txt_record.append(objects.Record(data='2'))
        records.append(
            objects.RecordSet(
                name=f'version.{pool.catalog_zone.catalog_zone_fqdn}',
                type='TXT',
                records=txt_record
            )
        )

        for z in zones:
            # If member zone is scheduled for deletion, do not include it in
            # catalog. Otherwise, zone poller will wait for zone's deletion on
            # secondary DNS servers, which will not happen since the zone is
            # still in catalog (deadlock).
            if z.action == 'DELETE':
                continue

            rs = objects.RecordList()
            rs.append(
                objects.Record(
                    data=z.name
                )
            )
            record = objects.RecordSet(
                name=f'{z.id}.zones.{soa_record.name}',
                type='PTR',
                records=rs
            )
            records.append(record)

        records.insert(0, soa_record)
        records.append(soa_record)

        return records



# Copyright 2015 Infoblox Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from urllib.parse import urlparse

from oslo_log import log as logging
from oslo_utils import importutils

from designate.backend import base
from designate import exceptions


infoblox_connector = importutils.try_import('infoblox_client.connector')
infoblox_exceptions = importutils.try_import('infoblox_client.exceptions')
infoblox_object_manager = importutils.try_import(
    'infoblox_client.object_manager'
)
infoblox_objects = importutils.try_import('infoblox_client.objects')


LOG = logging.getLogger(__name__)


class InfobloxBackend(base.Backend):
    """Provides a Designate Backend for Infoblox"""

    __backend_status__ = 'untested'
    __plugin_name__ = 'infoblox'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if not infoblox_connector:
            raise exceptions.Backend(
                'The infoblox-client library is not available'
            )

        wapi_host = self.options.get('wapi_host')
        wapi_version = self.options.get('wapi_version')
        wapi_url = self.options.get('wapi_url')

        self.multi_project = self.options.get('multi_tenant')
        self.dns_view = self.options.get('dns_view')
        self.network_view = self.options.get('network_view')
        self.ns_group = self.options.get('ns_group')

        if not wapi_host and wapi_url:
            wapi_host, wapi_version = self.parse_wapi_url(wapi_url)

        options = {
            'host': wapi_host,
            'username': self.options.get('username'),
            'password': self.options.get('password'),
            'http_pool_connections': self.options.get('http_pool_connections'),
            'http_pool_maxsize': self.options.get('http_pool_maxsize'),
            'wapi_version': wapi_version,
            'ssl_verify': self.options.get('sslverify'),
            'cert': self.options.get('cert'),
            'key': self.options.get('key'),
        }
        self.connection = infoblox_connector.Connector(options)
        self.infoblox = infoblox_object_manager.InfobloxObjectManager(
            self.connection
        )

        for master in self.masters:
            if master.port != 53:
                raise exceptions.ConfigurationError(
                    'Infoblox only supports mDNS instances on port 53'
                )

    def create_zone(self, context, zone):
        LOG.info('Create Zone %r', zone)

        dns_zone = zone['name'][0:-1]
        dns_view = self.dns_view
        project_id = context.project_id or zone.tenant_id

        if dns_zone.endswith('in-addr.arpa'):
            zone_format = 'IPV4'
        elif dns_zone.endswith('ip6.arpa'):
            zone_format = 'IPV6'
        else:
            zone_format = 'FORWARD'

        try:
            if self.is_multi_project:
                net_view = self.get_or_create_network_view(project_id)
                dns_view = self.get_or_create_dns_view(net_view)

            if not dns_view:
                raise exceptions.Backend(
                    'Unable to create zone. No DNS View found.'
                )

            self.infoblox.create_dns_zone(
                dns_zone=dns_zone,
                dns_view=dns_view,
                zone_format=zone_format,
                ns_group=self.ns_group,
            )
            self.restart_if_needed()
        except infoblox_exceptions.InfobloxException as e:
            raise exceptions.Backend(e)

    def delete_zone(self, context, zone, zone_params=None):
        LOG.info('Delete Zone %r', zone)

        dns_zone_fqdn = zone['name'][0:-1]
        dns_view = self.dns_view
        project_id = context.project_id or zone.tenant_id

        try:
            if self.is_multi_project:
                net_view = self.get_network_view(project_id)
                dns_view = self.get_or_create_dns_view(
                    net_view, create_if_missing=False
                )

            if not dns_view:
                raise exceptions.Backend(
                    'Unable to delete zone. No DNS View found.'
                )

            self.infoblox.delete_dns_zone(dns_view, dns_zone_fqdn)
            self.restart_if_needed()
        except infoblox_exceptions.InfobloxException as e:
            raise exceptions.Backend(e)

    @staticmethod
    def parse_wapi_url(wapi_url):
        url = urlparse(wapi_url)
        host = url.netloc
        wapi_version = None
        for path in url.path.split('/'):
            if path.startswith('v'):
                wapi_version = path.strip('v')
                break
        return host, wapi_version

    def get_network_view(self, project_id):
        network_views = self.connection.get_object(
            'networkview',
            return_fields=['name'],
            extattrs={'TenantID': {'value': project_id}}
        )
        network_view = None
        if network_views:
            network_view = network_views[0]['name']
        return network_view

    def get_or_create_network_view(self, project_id):
        network_view = self.get_network_view(project_id)
        if not network_view:
            network_view = self.infoblox.create_network_view(
                f'{self.network_view}.{project_id}',
                extattrs={'TenantID': {'value': project_id}}
            ).name
        return network_view

    def get_or_create_dns_view(self, net_view, create_if_missing=True):
        if not net_view:
            return None
        dns_view_name = f'{self.dns_view}.{net_view}'
        dns_view = infoblox_objects.DNSView.search(
            self.connection, name=dns_view_name, return_fields=['name'],
        )
        if not dns_view and create_if_missing:
            dns_view = self.infoblox.create_dns_view(
                self.network_view, dns_view_name
            )
        if not dns_view:
            return None
        return dns_view.name

    @property
    def is_multi_project(self):
        if not self.multi_project or self.multi_project == '0':
            return False
        return True

    def restart_if_needed(self):
        try:
            grid = infoblox_objects.Grid(self.connection)
            grid.fetch(only_ref=True)
            self.connection.call_func(
                'restartservices', grid._ref,
                {
                    'restart_option': 'RESTART_IF_NEEDED',
                    'mode': 'GROUPED',
                    'services': ['DNS'],
                }
            )
        except infoblox_exceptions.InfobloxException:
            LOG.warning('Unable to restart the infoblox dns service.')



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from oslo_log import versionutils
from oslo_policy import policy

from designate.common import constants
from designate.common.policies import base

DEPRECATED_REASON = """
The quota API now supports system scope and default roles.
"""

deprecated_get_quotas = policy.DeprecatedRule(
    name="get_quotas",
    check_str=base.RULE_ADMIN_OR_OWNER,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_set_quota = policy.DeprecatedRule(
    name="set_quota",
    check_str=base.RULE_ADMIN,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_reset_quotas = policy.DeprecatedRule(
    name="reset_quotas",
    check_str=base.RULE_ADMIN,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)

rules = [
    policy.DocumentedRuleDefault(
        name="get_quotas",
        check_str=base.SYSTEM_OR_PROJECT_READER_OR_ALL_TENANTS_READER,
        scope_types=[constants.PROJECT],
        description="View Current Project's Quotas.",
        operations=[
            {
                'path': '/v2/quotas',
                'method': 'GET'
            }
        ],
        deprecated_rule=deprecated_get_quotas
    ),
    policy.DocumentedRuleDefault(
        name="set_quota",
        check_str=base.SYSTEM_ADMIN,
        scope_types=[constants.PROJECT],
        description='Set Quotas.',
        operations=[
            {
                'path': '/v2/quotas/{project_id}',
                'method': 'PATCH'
            }
        ],
        deprecated_rule=deprecated_set_quota
    ),
    policy.DocumentedRuleDefault(
        name="reset_quotas",
        check_str=base.SYSTEM_ADMIN,
        scope_types=[constants.PROJECT],
        description='Reset Quotas.',
        operations=[
            {
                'path': '/v2/quotas/{project_id}',
                'method': 'DELETE'
            }
        ],
        deprecated_rule=deprecated_reset_quotas
    ),
]


def list_rules():
    return rules
